@@ -4,41 +4,38 @@
  *	Apply Function to Files
  */
 
+#include <fcntl.h>
 #include "fd.h"
 #include "term.h"
 #include "func.h"
 #include "kanji.h"
 
-#include <fcntl.h>
+#if	!MSDOS
 #include <sys/file.h>
-#include <sys/time.h>
 #include <sys/param.h>
-#include <sys/stat.h>
-
-#ifdef	USETIMEH
-#include <time.h>
-#endif
-
-#ifdef	USEUTIME
-#include <utime.h>
 #endif
 
 extern int filepos;
 extern reg_t *findregexp;
 extern int subwindow;
+extern int sizeinfo;
 
-static int judgecopy();
-static VOID showattr();
-static int touchfile();
+static long judgecopy __P_((char *, char *, time_t *, time_t *));
+static VOID showattr __P_((namelist *, u_short, char [][9], int y));
+static int touchfile __P_((char *, time_t, time_t));
 
 int copypolicy = 0;
 char *destpath = NULL;
 
-static u_short attrmode;
-static time_t attrtime;
+static u_short attrmode = 0;
+static time_t attrtime = 0;
+
+#ifndef	O_BINARY
+#define	O_BINARY	0
+#endif
 
 
-static int judgecopy(file, dest, atimep, mtimep)
+static long judgecopy(file, dest, atimep, mtimep)
 char *file, *dest;
 time_t *atimep, *mtimep;
 {
@@ -47,22 +44,26 @@ time_t *atimep, *mtimep;
 	int val[4];
 
 	strcpy(dest, destpath);
-	strcat(dest, "/");
+	strcat(dest, _SS_);
 	strcat(dest, file);
-	if (Xlstat(file, &status1) < 0) error(file);
-	*atimep = status1.st_atime;
-	*mtimep = status1.st_mtime;
+	if (Xlstat(file, &status1) < 0) {
+		warning(-1, file);
+		return(-1L);
+	}
+	if (atimep) *atimep = status1.st_atime;
+	if (mtimep) *mtimep = status1.st_mtime;
 	if (Xlstat(dest, &status2) < 0) {
-		if (errno != ENOENT) error(dest);
-		return((int)status1.st_mode);
+		if (errno == ENOENT) return((u_short)(status1.st_mode));
+		warning(-1, dest);
+		return(-1L);
 	}
 
 	if (!copypolicy || copypolicy == 2) {
 		locate(0, LCMDLINE);
 		putterm(l_clear);
-		putch('[');
+		putch2('[');
 		cp = SAMEF_K;
-		kanjiputs2(dest, n_column - strlen(cp) - 1, -1);
+		kanjiputs2(dest, n_column - (int)strlen(cp) - 1, -1);
 		kanjiputs(cp);
 	}
 	if (!copypolicy) {
@@ -80,32 +81,35 @@ time_t *atimep, *mtimep;
 	}
 	switch (copypolicy) {
 		case 1:
-			if (status1.st_mtime < status2.st_mtime) return(-1);
+			if (status1.st_mtime < status2.st_mtime) return(-1L);
 			break;
 		case 2:
-			cp = strrchr(dest, '/') + 1;
+			cp = strrchr(dest, _SC_) + 1;
 			do {
 				if (!(tmp = inputstr(NEWNM_K, 1,
-					-1, NULL, NULL))) return(-1);
+					-1, NULL, -1))) return(-1L);
 				strcpy(cp, tmp);
 				free(tmp);
 			} while (Xlstat(dest, &status2) >= 0
-			&& (putterm(t_bell) || 1));
-			if (errno != ENOENT) error(dest);
+			&& (putterm(t_bell), 1));
+			if (errno != ENOENT) {
+				warning(-1, dest);
+				return(-1L);
+			}
 			break;
 		case 3:
 			break;
 		default:
-			return(-1);
+			return(-1L);
 /*NOTREACHED*/
 			break;
 	}
-	return((int)status1.st_mode);
+	return((u_short)(status1.st_mode));
 }
 
 int _cpfile(src, dest, mode)
 char *src, *dest;
-int mode;
+u_short mode;
 {
 	char buf[BUFSIZ];
 	int i, fd1, fd2;
@@ -113,10 +117,11 @@ int mode;
 	if ((mode & S_IFMT) == S_IFLNK) {
 		if ((i = Xreadlink(src, buf, BUFSIZ)) < 0) return(-1);
 		buf[i] = '\0';
-		return(Xsymlink(buf, dest) < 0);
+		return(Xsymlink(buf, dest));
 	}
-	if ((fd1 = Xopen(src, O_RDONLY, mode)) < 0) return(-1);
-	if ((fd2 = Xopen(dest, O_WRONLY|O_CREAT|O_TRUNC, mode)) < 0) {
+	if ((fd1 = Xopen(src, O_RDONLY | O_BINARY, mode)) < 0) return(-1);
+	if ((fd2 = Xopen(dest,
+	O_WRONLY | O_BINARY | O_CREAT | O_TRUNC, mode)) < 0) {
 		Xclose(fd1);
 		return(-1);
 	}
@@ -124,13 +129,25 @@ int mode;
 	for (;;) {
 		while ((i = Xread(fd1, buf, BUFSIZ)) < 0 && errno == EINTR);
 		if (i < BUFSIZ) break;
-		if (Xwrite(fd2, buf, BUFSIZ) < 0 && errno != EINTR) error(dest);
+		if (Xwrite(fd2, buf, BUFSIZ) < 0 && errno != EINTR) {
+			i = -1;
+			break;
+		}
 	}
-	if (i < 0) error(src);
-	if (i > 0 && Xwrite(fd2, buf, i) < 0 && errno != EINTR) error(dest);
+	if (i > 0 && Xwrite(fd2, buf, i) < 0 && errno != EINTR) i = -1;
 
 	Xclose(fd2);
 	Xclose(fd1);
+
+	if (i < 0) {
+		i = errno;
+#if	MSDOS
+		if (!(mode & S_IWRITE)) Xchmod(dest, (mode | S_IWRITE));
+#endif
+		Xunlink(dest);
+		errno = i;
+		return(-1);
+	}
 	return(0);
 }
 
@@ -138,26 +155,47 @@ int cpfile(path)
 char *path;
 {
 	char dest[MAXPATHLEN + 1];
-	int mode, atime, mtime;
+	long mode;
+#if	MSDOS
+	time_t atime, mtime;
 
 	if ((mode = judgecopy(path, dest, &atime, &mtime)) < 0) return(0);
-	return(_cpfile(path, dest, mode));
+	if (_cpfile(path, dest, (u_short)mode) < 0) return(-1);
+	if (!(mode & S_IWRITE)) Xchmod(dest, (mode | S_IWRITE));
+	touchfile(dest, atime, mtime);
+	if (!(mode & S_IWRITE)) Xchmod(dest, mode);
+	return(0);
+#else
+
+	if ((mode = judgecopy(path, dest, NULL, NULL)) < 0) return(0);
+	return(_cpfile(path, dest, (u_short)mode));
+#endif
 }
 
 int mvfile(path)
 char *path;
 {
 	char dest[MAXPATHLEN + 1];
-	int mode, atime, mtime;
+	long mode;
+	time_t atime, mtime;
 
 	if ((mode = judgecopy(path, dest, &atime, &mtime)) < 0) return(0);
 	if (Xrename(path, dest) < 0) {
 		if (errno != EXDEV || (mode & S_IFMT) == S_IFDIR) return(-1);
+#if	MSDOS
+		if (!(mode & S_IWRITE)) Xchmod(path, (mode | S_IWRITE));
+		if (access2(path, W_OK) < 0) return(1);
+		if (_cpfile(path, dest, (mode | S_IWRITE)) < 0
+#else
 		if (access2(path, W_OK) < 0) return(1);
 		if (_cpfile(path, dest, mode) < 0
+#endif
 		|| Xunlink(path) < 0) return(-1);
 		if ((mode & S_IFMT) != S_IFLNK
 		&& touchfile(dest, atime, mtime) < 0) return(-1);
+#if	MSDOS
+		if (!(mode & S_IWRITE)) Xchmod(dest, mode);
+#endif
 	}
 	return(0);
 }
@@ -168,7 +206,7 @@ char *path;
 	char dest[MAXPATHLEN + 1];
 
 	strcpy(dest, destpath);
-	strcat(dest, "/");
+	strcat(dest, _SS_);
 	strcat(dest, path);
 	if (Xmkdir(dest, 0777) < 0 && errno != EEXIST) return(-1);
 	return(0);
@@ -179,16 +217,20 @@ char *path;
 {
 	char *cp;
 
-	if (cp = strrchr(path, '/')) cp++;
+	if ((cp = strrchr(path, _SC_))) cp++;
 	else cp = path;
 
 	if (regexp_exec(findregexp, cp)) {
+#if	MSDOS
+		if (!(strncmp(path, ".\\", 2))) path += 2;
+#else
 		if (!(strncmp(path, "./", 2))) path += 2;
+#endif
 		locate(0, LCMDLINE);
 		putterm(l_clear);
-		putch('[');
+		putch2('[');
 		kanjiputs2(path, n_column - 2, -1);
-		putch(']');
+		putch2(']');
 		if (yesno(FOUND_K)) {
 			destpath = strdup2(path);
 			return(-2);
@@ -202,7 +244,7 @@ char *path;
 {
 	char *cp;
 
-	if (cp = strrchr(path, '/')) cp++;
+	if ((cp = strrchr(path, _SC_))) cp++;
 	else cp = path;
 
 	if (regexp_exec(findregexp, cp)) {
@@ -221,7 +263,7 @@ char timestr[2][9];
 int y;
 {
 	struct tm *tm;
-	char buf[12];
+	char buf[WMODE + 1];
 
 	tm = localtime(&(listp -> st_mtim));
 
@@ -231,9 +273,9 @@ int y;
 	locate(0, ++y);
 	putterm(l_clear);
 	locate(n_column / 2 - 20, y);
-	putch('[');
+	putch2('[');
 	kanjiputs2(listp -> name, 16, 0);
-	putch(']');
+	putch2(']');
 	locate(n_column / 2 + 3, y);
 	kanjiputs(TOLD_K);
 	locate(n_column / 2 + 13, y);
@@ -245,30 +287,30 @@ int y;
 	kanjiputs(TMODE_K);
 	locate(n_column / 2, y);
 	putmode(buf, listp -> st_mode);
-	cputs(buf + 1);
+	cputs2(buf + 1);
 	locate(n_column / 2 + 10, y);
 	putmode(buf, mode);
-	cputs(buf + 1);
+	cputs2(buf + 1);
 
 	locate(0, ++y);
 	putterm(l_clear);
 	locate(n_column / 2 - 20, y);
 	kanjiputs(TDATE_K);
 	locate(n_column / 2, y);
-	cprintf("%02d-%02d-%02d",
+	cprintf2("%02d-%02d-%02d",
 		tm -> tm_year % 100, tm -> tm_mon + 1, tm -> tm_mday);
 	locate(n_column / 2 + 10, y);
-	cputs(timestr[0]);
+	cputs2(timestr[0]);
 
 	locate(0, ++y);
 	putterm(l_clear);
 	locate(n_column / 2 - 20, y);
 	kanjiputs(TTIME_K);
 	locate(n_column / 2, y);
-	cprintf("%02d:%02d:%02d",
+	cprintf2("%02d:%02d:%02d",
 		tm -> tm_hour, tm -> tm_min, tm -> tm_sec);
 	locate(n_column / 2 + 10, y);
-	cputs(timestr[1]);
+	cputs2(timestr[1]);
 
 	locate(0, ++y);
 	putterm(l_clear);
@@ -279,12 +321,14 @@ namelist *listp;
 u_short flag;
 {
 	struct tm *tm;
-	char buf[12], timestr[2][9];
+	char buf[WMODE + 1], timestr[2][9];
 	u_short tmp;
 	int i, ch, x, y, yy, ymin, ymax;
 
 	subwindow = 1;
-	getkey2(-1);
+#ifndef	_NOEDITMODE
+	Xgetkey(-1);
+#endif
 	yy = WHEADER;
 	while (n_line - yy < 7) yy--;
 
@@ -304,7 +348,7 @@ u_short flag;
 		tflush();
 
 		keyflush();
-		switch (ch = getkey2(0)) {
+		switch (ch = Xgetkey(0)) {
 			case K_UP:
 				if (y > ymin) y--;
 				else y = ymax;
@@ -344,7 +388,7 @@ u_short flag;
 			case '8':
 			case '9':
 				if (!y) break;
-				putch(ch);
+				putch2(ch);
 				timestr[y - 1][x] = ch;
 			case K_RIGHT:
 				if (y) {
@@ -358,7 +402,11 @@ u_short flag;
 						if (!((x + 1) % 3)) x++;
 					}
 				}
+#if	MSDOS
+				else if (x < 3) x++;
+#else
 				else if (x < 8) x++;
+#endif
 				break;
 			case K_LEFT:
 				if (y) {
@@ -379,6 +427,15 @@ u_short flag;
 				break;
 			case ' ':
 				if (y) break;
+#if	MSDOS
+				if (x == 2) break;
+				else if (x == 3) attrmode ^= S_ISVTX;
+				else {
+					tmp = 1;
+					for (i = 8; i > x; i--) tmp <<= 1;
+					attrmode ^= tmp;
+				}
+#else
 				tmp = 1;
 				for (i = 8; i > x; i--) tmp <<= 1;
 				if (!((x + 1) % 3) && (attrmode & tmp)) {
@@ -387,9 +444,10 @@ u_short flag;
 					else attrmode ^= (tmp << i);
 				}
 				attrmode ^= tmp;
+#endif
 				locate(n_column / 2 + 10, yy + y + 2);
 				putmode(buf, attrmode);
-				cputs(buf + 1);
+				cputs2(buf + 1);
 				break;
 			default:
 				break;
@@ -397,7 +455,9 @@ u_short flag;
 	} while (ch != ESC && ch != CR);
 
 	subwindow = 0;
-	getkey2(-1);
+#ifndef	_NOEDITMODE
+	Xgetkey(-1);
+#endif
 
 	if (ch == ESC) return(0);
 
@@ -446,10 +506,13 @@ char *path;
 	struct stat status;
 	u_short mode;
 
-	if (Xlstat(path, &status) < 0) error(path);
+	if (Xlstat(path, &status) < 0) return(-1);
 	if ((status.st_mode & S_IFMT) == S_IFLNK) return(1);
 
 	mode = (status.st_mode & S_IFMT) | (attrmode & ~S_IFMT);
+#if	MSDOS
+	if (!(status.st_mode & S_IWRITE)) Xchmod(path, (mode | S_IWRITE));
+#endif
 	if (touchfile(path, status.st_atime, attrtime) < 0
 	|| Xchmod(path, mode) < 0) return(-1);
 	return(0);
@@ -458,7 +521,7 @@ char *path;
 int applyfile(list, max, func, endmes)
 namelist *list;
 int max;
-int (*func)();
+int (*func)__P_((char *));
 char *endmes;
 {
 	int i, ret, tmp, old;
@@ -466,6 +529,11 @@ char *endmes;
 	ret = tmp = old = filepos;
 	helpbar();
 	for (filepos = 0; filepos < max; filepos++) {
+		if (kbhit2(0) && ((i = getkey2(0)) == cc_intr || i == ESC)) {
+			warning(0, INTR_K);
+			endmes = NULL;
+			break;
+		}
 		if (!ismark(&list[filepos])) continue;
 
 		movepos(list, max, old, 0);
@@ -484,16 +552,21 @@ char *endmes;
 
 int applydir(dir, funcf, funcd1, funcd2, endmes)
 char *dir;
-int (*funcf)();
-int (*funcd1)();
-int (*funcd2)();
+int (*funcf)__P_((char *));
+int (*funcd1)__P_((char *));
+int (*funcd2)__P_((char *));
 char *endmes;
 {
 	DIR *dirp;
 	struct dirent *dp;
 	struct stat status;
-	char *cp, path[MAXPATHLEN + 1];
-	int i;
+	char *cp, *fname, path[MAXPATHLEN + 1], **dirlist;
+	int i, ndir, max;
+
+	if (kbhit2(0) && ((i = getkey2(0)) == cc_intr || i == ESC)) {
+		warning(0, INTR_K);
+		return(-2);
+	}
 
 	if (!(dirp = Xopendir(dir))) {
 		warning(-1, dir);
@@ -502,36 +575,66 @@ char *endmes;
 
 	locate(0, LCMDLINE);
 	putterm(l_clear);
-	putch('[');
+	putch2('[');
+#if	MSDOS
+	cp = (strncmp(dir, ".\\", 2)) ? dir : dir + 2;
+#else
 	cp = (strncmp(dir, "./", 2)) ? dir : dir + 2;
+#endif
 	kanjiputs2(cp, n_column - 2, -1);
-	putch(']');
+	putch2(']');
 	tflush();
 
+	strcpy(path, dir);
+	fname = path + strlen(path);
+	*(fname++) = _SC_;
+
+	i = ndir = max = 0;
+	dirlist = NULL;
+	while ((dp = Xreaddir(dirp))) {
+		if (isdotdir(dp -> d_name)) continue;
+		strcpy(fname, dp -> d_name);
+
+		if (Xlstat(path, &status) >= 0
+		&& (status.st_mode & S_IFMT) == S_IFDIR) {
+			dirlist = (char **)addlist(dirlist, ndir,
+				&max, sizeof(char *));
+			dirlist[ndir++] = strdup2(dp -> d_name);
+		}
+	}
+	Xrewinddir(dirp);
+
 	if (funcd1 && (i = (*funcd1)(dir)) < 0) {
 		if (i == -1) warning(-1, dir);
 		return(i);
 	}
 
-	while (dp = Xreaddir(dirp)) {
-		if (!strcmp(dp -> d_name, ".")
-		|| !strcmp(dp -> d_name, "..")) continue;
+	max = ndir;
+	for (ndir = 0; ndir < max; ndir++) {
+		strcpy(fname, dirlist[ndir]);
+		free(dirlist[ndir]);
 
-		strcpy(path, dir);
-		strcat(path, "/");
-		strcat(path, dp -> d_name);
+		if ((i = applydir(path,
+			funcf, funcd1, funcd2, NULL)) < -1) return(i);
+		locate(0, LCMDLINE);
+		putterm(l_clear);
+		putch2('[');
+		kanjiputs2(cp, n_column - 2, -1);
+		putch2(']');
+		tflush();
+	}
+	free(dirlist);
+
+	while ((dp = Xreaddir(dirp))) {
+		if (kbhit2(0) && ((i = getkey2(0)) == cc_intr || i == ESC)) {
+			warning(0, INTR_K);
+			return(-2);
+		}
+		if (isdotdir(dp -> d_name)) continue;
+		strcpy(fname, dp -> d_name);
 
 		if (Xlstat(path, &status) < 0) warning(-1, path);
-		else if ((status.st_mode & S_IFMT) == S_IFDIR) {
-			if ((i = applydir(path,
-				funcf, funcd1, funcd2, NULL)) < -1) return(i);
-			locate(0, LCMDLINE);
-			putterm(l_clear);
-			putch('[');
-			kanjiputs2(cp, n_column - 2, -1);
-			putch(']');
-			tflush();
-		}
+		else if ((status.st_mode & S_IFMT) == S_IFDIR) continue;
 		else if ((i = (*funcf)(path)) < 0) {
 			if (i == -1) warning(-1, path);
 			else return(i);
