@@ -4,57 +4,88 @@
  *	Directory Browsing Module
  */
 
+#include <signal.h>
 #include "fd.h"
 #include "term.h"
 #include "func.h"
 #include "funcno.h"
 #include "kanji.h"
 
-#include <time.h>
-#include <signal.h>
+#if	MSDOS
+extern int getcurdrv __P_((VOID_A));
+extern int setcurdrv __P_((int));
+#else
 #include <sys/file.h>
 #include <sys/param.h>
-#include <sys/stat.h>
+#endif
 
 extern bindtable bindlist[];
 extern functable funclist[];
 extern char *findpattern;
-extern int sorttype;
+#ifndef	_NOWRITEFS
 extern int writefs;
-extern int minfilename;
+#endif
 extern char *curfilename;
+extern char *origpath;
 
-static VOID pathbar();
-static VOID stackbar();
-static char *putowner();
-static char *putgroup();
-static VOID calclocate();
-static int browsedir();
+static VOID pathbar __P_((VOID_A));
+static VOID stackbar __P_((VOID_A));
+#if	!MSDOS
+static char *putowner __P_((char *, uid_t));
+static char *putgroup __P_((char *, gid_t));
+#endif
+static VOID calclocate __P_((int));
+static int browsedir __P_((char *, char *));
 
-int columns;
-int filepos;
-int mark;
-int fnameofs;
-int dispmode;
-int sorton;
-int chgorder;
+int columns = 0;
+int minfilename = 0;
+int filepos = 0;
+int mark = 0;
+int fnameofs = 0;
+int dispmode = 0;
+int sorton = 0;
+int sorttype = 0;
+int chgorder = 0;
 int stackdepth = 0;
+int sizeinfo = 0;
+long marksize = 0;
+long totalsize = 0;
+long blocksize = 0;
 namelist filestack[MAXSTACK];
-char fullpath[MAXPATHLEN + 1];
-char *origpath;
+char fullpath[MAXPATHLEN + 1] = "";
 char *macrolist[MAXMACROTABLE];
 int maxmacro = 0;
+namelist *filelist = NULL;
+int maxfile = 0;
+int maxent = 0;
+int isearch = 0;
 char *helpindex[10] = {
+#ifdef	_NOTREE
+	"help", "eXec", "Copy", "Delete", "Rename",
+	"Sort", "Find", "Logdir", "Editor", "Unpack"
+#else
 	"Logdir", "eXec", "Copy", "Delete", "Rename",
 	"Sort", "Find", "Tree", "Editor", "Unpack"
+#endif
+};
+char typesymlist[] = "dbclsp";
+u_short typelist[] = {
+	S_IFDIR, S_IFBLK, S_IFCHR, S_IFLNK, S_IFSOCK, S_IFIFO
 };
 
-namelist *filelist;
-int maxfile;
-int maxent;
+static u_short modelist[] = {
+	S_IFDIR, S_IFLNK, S_IFSOCK, S_IFIFO
+};
+static char suffixlist[] = "/@=|";
+#ifndef	_NOCOLOR
+int ansicolor = 0;
+static u_char colorlist[] = {
+	ANSI_CYAN, ANSI_YELLOW, ANSI_MAGENTA, ANSI_RED
+};
+#endif
 
 
-static VOID pathbar()
+static VOID pathbar(VOID_A)
 {
 	char *path;
 
@@ -62,9 +93,9 @@ static VOID pathbar()
 
 	locate(0, LPATH);
 	putterm(l_clear);
-	putch(' ');
+	putch2(' ');
 	putterm(t_standout);
-	cputs("Path:");
+	cputs2("Path:");
 	putterm(end_standout);
 	kanjiputs2(path, n_column - 6, 0);
 	free(path);
@@ -72,33 +103,27 @@ static VOID pathbar()
 	tflush();
 }
 
-VOID helpbar()
+VOID helpbar(VOID_A)
 {
-	char *buf;
 	int i, j, width, len, ofs;
 
 	width = (n_column - 10) / 10 - 1;
 	ofs = (n_column - (width + 1) * 10 - 2) / 2;
-	buf = (char *)malloc2(width + 1);
 
 	locate(0, LHELP);
 	putterm(l_clear);
-	putch(isdisplnk(dispmode) ? 'S' : ' ');
-	putch(isdisptyp(dispmode) ? 'T' : ' ');
-	putch(ishidedot(dispmode) ? 'H' : ' ');
+	putch2(isdisplnk(dispmode) ? 'S' : ' ');
+	putch2(isdisptyp(dispmode) ? 'T' : ' ');
+	putch2(ishidedot(dispmode) ? 'H' : ' ');
 
 	for (i = 0; i < 10; i++) {
 		locate(ofs + (width + 1) * i + (i / 5) * 3, LHELP);
 		len = (width - strlen(helpindex[i])) / 2;
-		for (j = 0; j < len; j++) buf[j] = ' ';
-		strncpy2(buf + len, helpindex[i], width - len);
-		len = strlen(buf);
-		for (j = len; j < width; j++) buf[j] = ' ';
 		putterm(t_standout);
-		cputs(buf);
+		for (j = 0; j < len; j++) putch2(' ');
+		kanjiputs2(helpindex[i], width - len, 0);
 		putterm(end_standout);
 	}
-	free(buf);
 
 	tflush();
 }
@@ -106,26 +131,28 @@ VOID helpbar()
 VOID statusbar(max)
 int max;
 {
-	char *str[5];
+	char *str[6];
 	int width;
 
 	locate(0, LSTATUS);
 	putterm(l_clear);
-	putch(' ');
 
+	locate(CPAGE, LSTATUS);
 	putterm(t_standout);
-	cputs("Page:");
+	cputs2("Page:");
 	putterm(end_standout);
-	cprintf("%2d/%2d ",
+	cprintf2("%2d/%2d ",
 		filepos / FILEPERPAGE + 1, (max - 1) / FILEPERPAGE + 1);
 
+	locate(CMARK, LSTATUS);
 	putterm(t_standout);
-	cputs("Mark:");
+	cputs2("Mark:");
 	putterm(end_standout);
-	cprintf("%4d/%4d ", mark, max);
+	cprintf2("%4d/%4d ", mark, max);
 
+	locate(CSORT, LSTATUS);
 	putterm(t_standout);
-	cputs("Sort:");
+	cputs2("Sort:");
 	putterm(end_standout);
 
 	if (sorton & 7) {
@@ -133,19 +160,20 @@ int max;
 		str[1] = OEXT_K;
 		str[2] = OSIZE_K;
 		str[3] = ODATE_K;
+		str[4] = OLEN_K;
 		kanjiputs(str[(sorton & 7) - 1] + 3);
 
 		str[0] = OINC_K;
 		str[1] = ODEC_K;
-		putch('(');
+		putch2('(');
 		kanjiputs(str[sorton / 8] + 3);
-		putch(')');
+		putch2(')');
 	}
 	else kanjiputs(ORAW_K + 3);
 
 	locate(CFIND, LSTATUS);
 	putterm(t_standout);
-	cputs("Find:");
+	cputs2("Find:");
 	putterm(end_standout);
 	if (findpattern) {
 		width = n_column - (CFIND + 5);
@@ -155,21 +183,95 @@ int max;
 	tflush();
 }
 
-static VOID stackbar()
+static VOID stackbar(VOID_A)
 {
+#ifndef	_NOCOLOR
+	int j, x, color;
+#endif
 	int i, width;
 
 	width = n_column / MAXSTACK;
 
 	locate(0, LSTACK);
+#ifndef	_NOCOLOR
+	if (ansicolor == 2) chgcolor(ANSI_BLACK, 1);
+	x = 0;
+#endif
 	putterm(l_clear);
 
 	for (i = 0; i < stackdepth; i++) {
+#ifndef	_NOCOLOR
+		if (ansicolor == 2) for (; x < width * i + 1; x++) putch2(' ');
+		else
+#endif
 		locate(width * i + 1, LSTACK);
+#ifndef	_NOCOLOR
+		if (!isread(&filestack[i])) color = ANSI_BLUE;
+		else if (!iswrite(&filestack[i])) color = ANSI_GREEN;
+		else {
+			for (j = 0; j < sizeof(modelist) / sizeof(u_short); j++)
+				if ((filestack[i].st_mode & S_IFMT)
+				== modelist[j]) break;
+			if (j < sizeof(modelist) / sizeof(u_short))
+				color = colorlist[j];
+			else if (ansicolor == 3) color = ANSI_BLACK;
+			else color = ANSI_WHITE;
+		}
+		if (ansicolor) chgcolor(color, 1);
+		else
+#endif
 		putterm(t_standout);
 		kanjiputs2(filestack[i].name, width - 2, 0);
+#ifndef	_NOCOLOR
+		x += width - 2;
+		if (ansicolor == 2) chgcolor(ANSI_BLACK, 1);
+		else if (ansicolor) putterms(t_normal);
+		else
+#endif
 		putterm(end_standout);
 	}
+#ifndef	_NOCOLOR
+	if (ansicolor == 2) {
+		for (; x < n_column; x++) putch2(' ');
+		putterm(t_normal);
+	}
+#endif
+
+	tflush();
+}
+
+VOID sizebar(VOID_A)
+{
+	char buf[16];
+	long total, fre;
+
+	if (!sizeinfo) return;
+
+	locate(0, LSIZE);
+	putterm(l_clear);
+
+	locate(CSIZE, LSIZE);
+	putterm(t_standout);
+	cputs2("Size:");
+	putterm(end_standout);
+	cprintf2("%14.14s/", inscomma(buf, marksize, 3));
+	cprintf2("%14.14s ", inscomma(buf, totalsize, 3));
+
+	getinfofs(".", &total, &fre);
+
+	locate(CTOTAL, LSIZE);
+	putterm(t_standout);
+	cputs2("Total:");
+	putterm(end_standout);
+	if (total < 0) cprintf2("%15.15s ", "?");
+	else cprintf2("%12.12s KB ", inscomma(buf, total, 3));
+
+	locate(CFREE, LSIZE);
+	putterm(t_standout);
+	cputs2("Free:");
+	putterm(end_standout);
+	if (fre < 0) cprintf2("%15.15s ", "?");
+	else cprintf2("%12.12s KB ", inscomma(buf, fre, 3));
 
 	tflush();
 }
@@ -178,32 +280,19 @@ char *putmode(buf, mode)
 char *buf;
 u_short mode;
 {
-	switch(mode & S_IFMT) {
-		case S_IFDIR:
-			buf[0] = 'd';
-			break;
-		case S_IFBLK:
-			buf[0] = 'b';
-			break;
-		case S_IFCHR:
-			buf[0] = 'c';
-			break;
-		case S_IFLNK:
-			buf[0] = 'l';
-			break;
-		case S_IFSOCK:
-			buf[0] = 's';
-			break;
-		case S_IFIFO:
-			buf[0] = 'p';
-			break;
-		default:
-			buf[0] = '-';
-			break;
-	}
+	int i;
 
+	for (i = 0; i < sizeof(typelist) / sizeof(u_short); i++)
+		if ((mode & S_IFMT) == typelist[i]) break;
+	buf[0] = (i < sizeof(typelist) / sizeof(u_short))
+		? typesymlist[i] : '-';
 	buf[1] = (mode & S_IRUSR) ? 'r' : '-';
 	buf[2] = (mode & S_IWUSR) ? 'w' : '-';
+#if	MSDOS
+	buf[3] = (mode & S_IXUSR) ? 'x' : '-';
+	buf[4] = (mode & S_ISVTX) ? 'a' : '-';
+	buf[5] = '\0';
+#else
 	buf[3] = (mode & S_ISUID) ? 's' : (mode & S_IXUSR) ? 'x' : '-';
 	buf[4] = (mode & S_IRGRP) ? 'r' : '-';
 	buf[5] = (mode & S_IWGRP) ? 'w' : '-';
@@ -212,18 +301,20 @@ u_short mode;
 	buf[8] = (mode & S_IWOTH) ? 'w' : '-';
 	buf[9] = (mode & S_ISVTX) ? 't' : (mode & S_IXOTH) ? 'x' : '-';
 	buf[10] = '\0';
+#endif
 
 	return(buf);
 }
 
+#if	!MSDOS
 static char *putowner(buf, uid)
 char *buf;
 uid_t uid;
 {
 	char *str;
 
-	if (str = getpwuid2(uid)) strncpy3(buf, str, WOWNER, 0);
-	else sprintf(buf, "%-*d", WOWNER, uid);
+	if ((str = getpwuid2(uid))) strncpy3(buf, str, WOWNER, 0);
+	else sprintf(buf, "%-*u", WOWNER, uid);
 
 	return(buf);
 }
@@ -234,11 +325,12 @@ gid_t gid;
 {
 	char *str;
 
-	if (str = getgrgid2(gid)) strncpy3(buf, str, WGROUP, 0);
-	else sprintf(buf, "%-*d", WGROUP, gid);
+	if ((str = getgrgid2(gid))) strncpy3(buf, str, WGROUP, 0);
+	else sprintf(buf, "%-*u", WGROUP, gid);
 
 	return(buf);
 }
+#endif	/* !MSDOS */
 
 VOID infobar(list, no)
 namelist *list;
@@ -251,6 +343,7 @@ int no;
 	locate(0, LINFO);
 
 	if (list[no].st_nlink < 0) {
+		putterm(l_clear);
 		kanjiputs(list[no].name);
 		tflush();
 		return;
@@ -265,17 +358,19 @@ int no;
 	sprintf(buf + len, " %2d ", list[no].st_nlink);
 	len += 4;
 
+#if	!MSDOS
 	putowner(buf + len, list[no].st_uid);
 	len += WOWNER;
 	strcpy(buf + (len++), " ");
 	putgroup(buf + len, list[no].st_gid);
 	len += WGROUP;
 	strcpy(buf + (len++), " ");
+#endif
 
 	if (isdev(&list[no])) sprintf(buf + len, "%3u, %3u ",
 		((unsigned)(list[no].st_size) >> 8) & 0xff,
 		(unsigned)(list[no].st_size) & 0xff);
-	else sprintf(buf + len, "%8d ", (unsigned)(list[no].st_size));
+	else sprintf(buf + len, "%8ld ", (long)(list[no].st_size));
 	len = strlen(buf);
 
 	sprintf(buf + len, "%02d-%02d-%02d %02d:%02d ",
@@ -285,6 +380,7 @@ int no;
 	width = n_lastcolumn - len;
 	strncpy3(buf + len, list[no].name, width, fnameofs);
 
+#if	!MSDOS
 	if (islink(&list[no])) {
 		len = strlen(buf);
 		while (buf[--len] == ' ');
@@ -299,12 +395,13 @@ int no;
 			Xreadlink(list[no].name, buf + len, width);
 		}
 	}
+#endif
 
 	kanjiputs(buf);
 	tflush();
 }
 
-VOID waitmes()
+VOID waitmes(VOID_A)
 {
 	helpbar();
 	locate(0, LMESLINE);
@@ -321,14 +418,26 @@ int i;
 	i %= FILEPERPAGE;
 	x = (i / FILEPERLOW) * (n_column / FILEPERLINE);
 	y = i % FILEPERLOW;
+#ifndef	_NOCOLOR
+	if (ansicolor == 2) {
+		chgcolor(ANSI_BLACK, 1);
+		locate(x, y + WHEADER);
+		putch2(' ');
+	}
+	else
+#endif
 	locate(x + 1, y + WHEADER);
 }
 
+#if	MSDOS
+#define	WIDTH1	(WMODE + 1 + WSECOND + 1)
+#else
 #define	WIDTH1	(WOWNER + 1 + WGROUP + 1 + WMODE + 1 + WSECOND + 1)
+#endif
 #define	WIDTH2	(WTIME + 1 + WDATE + 1)
 #define	WIDTH3	(WSIZE + 1)
 
-int calcwidth()
+int calcwidth(VOID_A)
 {
 	int width;
 
@@ -345,10 +454,13 @@ int no, standout;
 {
 	char buf[MAXLINESTR + 1];
 	struct tm *tm;
-	int i, len, width;
+	int i, j, len, width;
+#ifndef	_NOCOLOR
+	int color;
+#endif
 
 	calclocate(no);
-	putch(ismark(&list[no]) ? '*' : ' ');
+	putch2(ismark(&list[no]) ? '*' : ' ');
 
 	if (standout < 0 && stable_standout) {
 		putterm(end_standout);
@@ -359,44 +471,48 @@ int no, standout;
 	width = calcwidth();
 	i = (standout && fnameofs > 0) ? fnameofs : 0;
 	i = strncpy3(buf, list[no].name, width, i);
-	if (isdisptyp(dispmode) && i < width) {
-		switch (list[no].st_mode & S_IFMT) {
-			case S_IFDIR:
-				buf[i] = '/';
-				break;
-			case S_IFLNK:
-				buf[i] = '@';
-				break;
-			case S_IFSOCK:
-				buf[i] = '=';
-				break;
-			case S_IFIFO:
-				buf[i] = '|';
-				break;
-			case S_IFREG:
-				if (list[no].st_mode &
-				(S_IXUSR | S_IXGRP | S_IXOTH))
-					buf[i] = '*';
-				break;
-			default:
-				break;
-		}
+#ifdef	_NOCOLOR
+	if (isdisptyp(dispmode) && i < width)
+#endif
+	{
+		for (j = 0; j < sizeof(modelist) / sizeof(u_short); j++)
+			if ((list[no].st_mode & S_IFMT) == modelist[j]) break;
+#ifndef	_NOCOLOR
 	}
+	if (isdisptyp(dispmode) && i < width) {
+#endif
+		if (j < sizeof(modelist) / sizeof(u_short))
+			buf[i] = suffixlist[j];
+		else if ((list[no].st_mode & S_IFMT) == S_IFREG
+		&& (list[no].st_mode & (S_IXUSR | S_IXGRP | S_IXOTH)))
+			buf[i] = '*';
+	}
+#ifndef	_NOCOLOR
+	if (!isread(&list[no])) color = ANSI_BLUE;
+	else if (!iswrite(&list[no])) color = ANSI_GREEN;
+	else if (j < sizeof(modelist) / sizeof(u_short)) color = colorlist[j];
+	else if (ansicolor == 3) color = ANSI_BLACK;
+	else color = ANSI_WHITE;
+#endif
 	len = width;
 	width = (n_column / columns) - 2 - 1;
 
+	tm = NULL;
 	if (columns < 5 && len + WIDTH3 <= width) {
 		if (isdir(&list[no]))
 			sprintf(buf + len, " %*.*s", WSIZE, WSIZE, "<DIR>");
-		else if (dospath("", NULL)
-		&& (list[no].st_mode & S_IFMT) == S_IFIFO)
+		else if (
+#if	!MSDOS && !defined (_NODOSDRIVE)
+		dospath("", NULL) &&
+#endif
+		(list[no].st_mode & S_IFMT) == S_IFIFO)
 			sprintf(buf + len, " %*.*s", WSIZE, WSIZE, "<VOL>");
 		else if (isdev(&list[no])) sprintf(buf + len, " %*u,%*u",
 			WSIZE / 2, ((unsigned)(list[no].st_size) >> 8) & 0xff,
 			WSIZE - (WSIZE / 2) - 1,
 			(unsigned)(list[no].st_size) & 0xff);
-		else sprintf(buf + len, " %*d",
-			WSIZE, (unsigned)(list[no].st_size));
+		else sprintf(buf + len, " %*ld",
+			WSIZE, (long)(list[no].st_size));
 		if (strlen(buf + len) > WSIZE + 1)
 			sprintf(buf + len, " %*.*s", WSIZE, WSIZE, "OVERFLOW");
 		len += WIDTH3;
@@ -409,21 +525,36 @@ int no, standout;
 		len += WIDTH2;
 	}
 	if (columns < 2 && len + WIDTH1 <= width) {
+		if (!tm) tm = localtime(&list[no].st_mtim);
 		sprintf(buf + len, ":%02d", tm -> tm_sec);
 		len += 1 + WSECOND;
 		buf[len++] = ' ';
+#if	!MSDOS
 		putowner(buf + len, list[no].st_uid);
 		len += WOWNER;
 		buf[len++] = ' ';
 		putgroup(buf + len, list[no].st_gid);
 		len += WGROUP;
 		buf[len++] = ' ';
+#endif
 		putmode(buf + len, (!isdisplnk(dispmode) && islink(&list[no])) ?
 			(S_IFLNK | 0777) : list[no].st_mode);
 	}
 
+#ifndef	_NOCOLOR
+	if (ansicolor) chgcolor(color, standout > 0);
+	else
+#endif
 	if (standout > 0) putterm(t_standout);
 	kanjiputs(buf);
+#ifndef	_NOCOLOR
+	if (ansicolor == 2) {
+		chgcolor(ANSI_BLACK, 1);
+		putch2(' ');
+	}
+	if (ansicolor) putterms(t_normal);
+	else
+#endif
 	if (standout > 0) putterm(end_standout);
 	tflush();
 }
@@ -447,7 +578,7 @@ char *def;
 		putterm(t_standout);
 		cp = NOFIL_K;
 		if (i > strlen(cp)) kanjiputs2(cp, i, 0);
-		else cprintf("%-*.*s", i, i, "No Files");
+		else cprintf2("%-*.*s", i, i, "No Files");
 		putterm(end_standout);
 		tflush();
 		return(0);
@@ -462,7 +593,7 @@ char *def;
 			count = 0;
 			start = i;
 		}
-		if (!strcmp(def, list[i].name)) {
+		if (!strpathcmp(def, list[i].name)) {
 			ret = i;
 			break;
 		}
@@ -473,7 +604,7 @@ char *def;
 	}
 
 	locate(6, LSTATUS);
-	cprintf("%2d", start / FILEPERPAGE + 1);
+	cprintf2("%2d", start / FILEPERPAGE + 1);
 
 	for (i = start, count = 0; i < max; i++, count++) {
 		if (count >= FILEPERPAGE) break;
@@ -504,6 +635,7 @@ VOID rewritefile(all)
 int all;
 {
 	pathbar();
+	sizebar();
 	if (all) {
 		helpbar();
 		infobar(filelist, filepos);
@@ -515,6 +647,84 @@ int all;
 	tflush();
 }
 
+int searchmove(list, max, ch, buf)
+namelist *list;
+int max, ch;
+char *buf;
+{
+	char *str[4];
+	int i, n, s, pos, len;
+
+	if (isearch > 0) {
+		n = isearch;
+		s = 1;
+	}
+	else {
+		n = -isearch;
+		s = -1;
+	}
+	len = strlen(buf);
+	pos = filepos;
+
+	for (i = 0; i < MAXBINDTABLE && bindlist[i].key >= 0; i++)
+		if (ch == (int)(bindlist[i].key)) break;
+	if (ch == K_BS) {
+		if (!len) pos = -1;
+		else buf[--len] = '\0';
+	}
+	else if (len && bindlist[i].f_func == SEARCH_FORW) {
+		if (n > 2) pos++;
+		else if (n > 1) pos = 0;
+		s = 1;
+	}
+	else if (len && bindlist[i].f_func == SEARCH_BACK) {
+		if (n > 2) pos--;
+		else if (n > 1) pos = max - 1;
+		s = -1;
+	}
+	else {
+		if (n == 1) buf[len = 0] = '\0';
+		if (ch < ' ' || ch == C_DEL || ch >= K_MIN) pos = -1;
+		else if (len < MAXNAMLEN - 1) {
+			buf[len++] = ch;
+			buf[len] = '\0';
+		}
+	}
+
+	if (pos < 0 || pos >= max) {
+		isearch = 0;
+		helpbar();
+		if (ch == CR) return(-1);
+		else if (ch != ESC) putterm(t_bell);
+		return(0);
+	}
+
+	locate(0, LHELP);
+	str[0] = SEAF_K;
+	str[1] = SEAFF_K;
+	str[2] = SEAB_K;
+	str[3] = SEABF_K;
+
+	i = 0;
+	for (;;) {
+		if (!strnpathcmp(list[pos].name, buf, len)) {
+			i = 1;
+			break;
+		}
+		pos += s;
+		if (pos < 0 || pos >= max) break;
+	}
+
+	putterm(t_standout);
+	len = kanjiputs(str[2 - s - i]);
+	putterm(end_standout);
+	kanjiputs2(buf, n_column - len - 1, 0);
+	if (i) filepos = pos;
+	else if (n != 2 && ch != K_BS) putterm(t_bell);
+	isearch = s * (i + 2);
+	return(i);
+}
+
 static int browsedir(file, def)
 char *file, *def;
 {
@@ -522,10 +732,17 @@ char *file, *def;
 	struct dirent *dp;
 	reg_t *re;
 	u_char fstat;
-	char *cp;
+	char *cp, buf[MAXNAMLEN + 1];
 	int ch, i, no, old;
 
+#ifndef	_NOCOLOR
+	if (ansicolor) putterms(t_normal);
+#endif
+
 	maxfile = mark = 0;
+	totalsize = marksize = 0;
+	buf[0] = '\0';
+	blocksize = getblocksize(".");
 	chgorder = 0;
 	if (sorttype < 100) sorton = sorttype;
 
@@ -539,10 +756,9 @@ char *file, *def;
 		re = regexp_init(cp);
 		free(cp);
 	}
-	while (dp = searchdir(dirp, re)) {
+	while ((dp = searchdir(dirp, re))) {
 		if (ishidedot(dispmode) && *(dp -> d_name) == '.'
-		&& strcmp(dp -> d_name, ".")
-		&& strcmp(dp -> d_name, "..")) continue;
+		&& !isdotdir(dp -> d_name)) continue;
 		strcpy(file, dp -> d_name);
 		for (i = 0; i < stackdepth; i++)
 			if (!strcmp(file, filestack[i].name)) break;
@@ -553,6 +769,10 @@ char *file, *def;
 		if (getstatus(filelist, maxfile, file) < 0) continue;
 		filelist[maxfile].name = strdup2(file);
 		filelist[maxfile].ent = maxfile;
+		if (!isdir(&filelist[maxfile]))
+			totalsize += 
+				((filelist[maxfile].st_size + blocksize - 1)
+				/ blocksize) * blocksize;
 		maxfile++;
 	}
 	Xclosedir(dirp);
@@ -573,6 +793,7 @@ char *file, *def;
 	}
 	title();
 	pathbar();
+	sizebar();
 	statusbar(maxfile);
 	stackbar();
 
@@ -585,17 +806,28 @@ char *file, *def;
 		if (no) movepos(filelist, maxfile, old, fstat);
 		locate(0, 0);
 		tflush();
-		getkey2(-1);
-		ch = getkey2(SIGALRM);
-		getkey2(-1);
+#ifdef	_NOEDITMODE
+		ch = Xgetkey(SIGALRM);
+#else
+		Xgetkey(-1);
+		ch = Xgetkey(SIGALRM);
+		Xgetkey(-1);
+#endif
 
 		old = filepos;
+		if (isearch) {
+			no = searchmove(filelist, maxfile, ch, buf);
+			if (no >= 0) {
+				fnameofs = 0;
+				continue;
+			}
+		}
 		for (i = 0; i < MAXBINDTABLE && bindlist[i].key >= 0; i++)
 			if (ch == (int)(bindlist[i].key)) break;
 		no = (bindlist[i].d_func < 255 && isdir(&filelist[filepos])) ?
 			(int)(bindlist[i].d_func) : (int)(bindlist[i].f_func);
 		fstat = (no <= NO_OPERATION) ? funclist[no].stat
-			: (KILLSTK | REWRITE);
+			: (KILLSTK | RELIST);
 
 		if (fstat & KILLSTK) {
 			if (stackdepth > 0) {
@@ -614,9 +846,9 @@ char *file, *def;
 					filelist[filepos].name);
 				stackbar();
 			}
-#if (WRITEFS < 1)
+#ifndef	_NOWRITEFS
 			else if (chgorder && writefs < 1 && no != WRITE_DIR
-			&& (i = writablefs(".")) > 0 && underhome() > 0) {
+			&& (i = writablefs(".")) > 0 && underhome(NULL) > 0) {
 				chgorder = 0;
 				if (yesno(WRTOK_K))
 					arrangedir(filelist, maxfile, i);
@@ -627,12 +859,12 @@ char *file, *def;
 		curfilename = filelist[filepos].name;
 		if (maxfile <= 0 && !(fstat & NO_FILE)) no = 0;
 		else if (no <= NO_OPERATION)
-			no = (*funclist[no].func)(filelist, &maxfile);
+			no = (*funclist[no].func)(filelist, &maxfile, NULL);
 		else {
-			execmacro(macrolist[no - NO_OPERATION - 1],
+			no = execusercomm(macrolist[no - NO_OPERATION - 1],
 				filelist[filepos].name,
-				filelist, maxfile, 0, 0);
-			no = 4;
+				filelist, &maxfile, 0, 0);
+			if (no < 0) no = 0;
 		}
 
 		if (no < 0) break;
@@ -651,6 +883,7 @@ char *file, *def;
 		if ((fstat & RELIST) == RELIST) {
 			title();
 			pathbar();
+			sizebar();
 			statusbar(maxfile);
 			stackbar();
 			helpbar();
@@ -668,28 +901,50 @@ char *cur;
 	char *cp, *def;
 	int i, ischgdir;
 
-	origpath = getwd2();
-	strcpy(fullpath, origpath);
-
 	findpattern = def = NULL;
 	if (cur) {
-		cp = evalpath(strdup2(cur));
-		if (_dospath(cp) || *cp == '/') *fullpath = '\0';
-		else strcat(fullpath, "/");
+		cp = evalpath(strdup2(cur), 1);
+#if	MSDOS
+		if (_dospath(cp)) {
+			if (setcurdrv(*cp) < 0) fullpath[0] = getcurdrv();
+			else {
+				def = getwd2();
+				strcpy(fullpath, def);
+				def = NULL;
+			}
+			for (i = 2; cp[i]; i++) cp[i - 2] = cp[i];
+			cp [i - 2] = '\0';
+		}
+		if (*cp == _SC_) fullpath[2] = '\0';
+		strcat(fullpath, _SS_);
+#else	/* !MSDOS */
+		if (
+#ifndef	_NODOSDRIVE
+		_dospath(cp) ||
+#endif
+		*cp == _SC_) *fullpath = '\0';
+		else strcat(fullpath, _SS_);
+#endif	/* !MSDOS */
 		if (_chdir2(cp) >= 0) {
 			strcat(fullpath, cp);
 			free(cp);
-			cp = fullpath + strlen(fullpath) - 1;
+			cp = fullpath + (int)strlen(fullpath) - 1;
 		}
 		else {
-			def = strrchr(cp, '/');
+			def = strrchr(cp, _SC_);
+#if	!MSDOS && !defined (_NODOSDRIVE)
 			if (!def && _dospath(cp)) def = &cp[2];
+#endif
 			if (def) {
 				i = *def;
 				*def = '\0';
 				if (def == cp) {
-					if (_chdir2("/") < 0) error("/");
-					else strcpy(fullpath, "/");
+					if (_chdir2(_SS_) < 0) error(_SS_);
+#if	MSDOS
+					strcpy(fullpath + 2, _SS_);
+#else
+					strcpy(fullpath, _SS_);
+#endif
 				}
 				else if (_chdir2(cp) >= 0) strcat(fullpath, cp);
 				else {
@@ -697,7 +952,7 @@ char *cur;
 					def = NULL;
 					strcpy(fullpath, origpath);
 				}
-				if (def && (*def = i) == '/') def++;
+				if (def && (*def = i) == _SC_) def++;
 			}
 			else def = cp;
 
@@ -708,6 +963,7 @@ char *cur;
 			free(cp);
 		}
 		realpath2(fullpath, fullpath);
+		entryhist(1, fullpath, 1);
 	}
 
 	filelist = NULL;
@@ -717,7 +973,7 @@ char *cur;
 
 	for (;;) {
 		if (!def && !strcmp(file, "..")) {
-			if (cp = strrchr(fullpath, '/')) cp++;
+			if ((cp = strrchr(fullpath, _SC_))) cp++;
 			else cp = fullpath;
 			strcpy(prev, cp);
 			if (!*prev) {
@@ -747,7 +1003,5 @@ char *cur;
 		}
 	}
 
-	free(filelist);
-	if (_chdir2(origpath) < 0) error(origpath);
-	free(origpath);
+	if (filelist) free(filelist);
 }
