@@ -4,19 +4,31 @@
  *	Information Module
  */
 
+#include <ctype.h>
 #include "fd.h"
 #include "term.h"
 #include "funcno.h"
 #include "kanji.h"
 
-#include <sys/param.h>
-#include <sys/file.h>
-#include <sys/stat.h>
+#ifdef	NOERRNO
+extern int errno;
+#endif
 
-#ifdef	SYSDIRH
+#ifdef	USESYSDIRH
 #include <sys/dir.h>
 #endif
 
+#if	MSDOS
+#undef	USEVFSH
+#undef	USEMNTENTH
+#include <dos.h>
+#include <errno.h>
+#include "unixemu.h"
+extern int supportLFN __P_((char *));
+#else
+#include <sys/param.h>
+#include <sys/file.h>
+#endif
 
 #ifdef	USEMNTENTH
 #include <mntent.h>
@@ -30,11 +42,14 @@ typedef struct mntent	mnt_t;
 typedef struct mnttab	mnt_t;
 #define	setmntent		fopen
 #define	getmntent2(fp, mntp)	(getmntent(fp, mntp) ? NULL : mntp)
-#define	hasmntopt(mntp, opt)	strstr((mntp) -> mnt_mntopts, opt)
+#define	hasmntopt(mntp, opt)	strstr2((mntp) -> mnt_mntopts, opt)
 #define	endmntent		fclose
 #define	mnt_dir		mnt_mountp
 #define	mnt_fsname	mnt_special
 #define	mnt_type	mnt_fstype
+#ifndef	mnt_opts
+#define	mnt_opts	mnt_mntopts
+#endif
 #endif	/* USEMNTTABH */
 
 #ifdef	USEMNTCTL
@@ -42,18 +57,18 @@ typedef struct mnttab	mnt_t;
 #include <sys/vfs.h>
 #include <sys/mntctl.h>
 #include <sys/vmount.h>
-#endif
+#endif	/* USEMNTCTL */
 
-#if (defined (USEGETFSSTAT) || defined (USEMNTINFOR) || defined (USEMNTINFO))\
-&& !defined (USEMOUNTH) && !defined (USEFSDATA)
+#if !defined (USEMOUNTH) && !defined (USEFSDATA)\
+&& (defined (USEGETFSSTAT) || defined (USEMNTINFOR) || defined (USEMNTINFO))
 #include <sys/mount.h>
 #endif
 
-#if defined (USEGETFSSTAT) || defined (USEGETMNT)
+#if	defined (USEGETFSSTAT) || defined (USEGETMNT)
 #include <sys/fs_types.h>
 #endif
 
-#if defined (USEGETFSSTAT) || defined (USEMNTCTL)\
+#if	defined (USEGETFSSTAT) || defined (USEMNTCTL)\
 || defined (USEMNTINFOR) || defined (USEMNTINFO) || defined (USEGETMNT)
 typedef struct _mnt_t {
 	char *mnt_fsname;
@@ -61,12 +76,16 @@ typedef struct _mnt_t {
 	char *mnt_type;
 	char *mnt_opts;
 } mnt_t;
-static FILE *setmntent();
-static mnt_t *getmntent2();
-#define	hasmntopt(mntp, opt)	strstr((mntp) -> mnt_opts, opt)
+static FILE *setmntent __P_((char *, char *));
+static mnt_t *getmntent2 __P_((FILE *, mnt_t *));
+#define	hasmntopt(mntp, opt)	strstr2((mntp) -> mnt_opts, opt)
+#if	defined(USEMNTINFO) || defined(USEGETMNT)
+#define	endmntent(fp)
+#else
 #define	endmntent(fp)		{ if (fp) free(fp); }
-static int mnt_ptr;
-static int mnt_size;
+#endif
+static int mnt_ptr = 0;
+static int mnt_size = 0;
 #endif
 
 #ifdef	USEGETFSENT
@@ -74,13 +93,23 @@ static int mnt_size;
 typedef struct fstab	mnt_t;
 #define	setmntent(file, mode)	(FILE *)(setfsent(), NULL)
 #define	getmntent2(fp, mntp)	getfsent()
-#define	hasmntopt(mntp, opt)	strstr((mntp) -> fs_mntops, opt)
+#define	hasmntopt(mntp, opt)	strstr2((mntp) -> fs_mntops, opt)
 #define	endmntent(fp)		endfsent()
 #define	mnt_dir		fs_file
 #define	mnt_fsname	fs_spec
 #define	mnt_type	fs_vfstype
 #endif	/* USEGETFSENT */
 
+#if	MSDOS
+typedef struct _mnt_t {
+	char *mnt_fsname;
+	char mnt_dir[4];
+	char *mnt_type;
+	char *mnt_opts;
+} mnt_t;
+#define	hasmntopt(mntp, opt)	strstr2((mntp) -> mnt_opts, opt)
+#endif
+
 
 #ifdef	USESTATVFSH
 #include <sys/statvfs.h>
@@ -121,11 +150,11 @@ typedef struct fs_data	statfs_t;
 #define	f_bsize		f_fsize
 #endif
 
-#if defined (USESTATFSH) || defined (USEVFSH) || defined (USEMOUNTH)
-# if (STATFSARGS >= 4)
+#if	defined (USESTATFSH) || defined (USEVFSH) || defined (USEMOUNTH)
+# if	(STATFSARGS >= 4)
 # define	statfs2(path, buf)	statfs(path, buf, sizeof(statfs_t), 0)
 # else
-#  if (STATFSARGS == 3)
+#  if	(STATFSARGS == 3)
 #  define	statfs2(path, buf)	statfs(path, buf, sizeof(statfs_t))
 #  else
 #  define	statfs2			statfs
@@ -133,15 +162,40 @@ typedef struct fs_data	statfs_t;
 # endif
 #endif
 
-extern VOID error();
-extern char *getwd2();
-extern VOID warning();
-extern char *mesconv();
-extern int kanjiputs2();
+#if	MSDOS
+typedef struct _statfs_t {
+	long	f_bsize;
+	long	f_blocks;
+	long	f_bfree;
+	long	f_bavail;
+	long	f_files;
+} statfs_t;
+static int statfs2 __P_((char *, statfs_t *));
+#define	blocksize(fs)	(fs).f_bsize
+#endif
+
+extern VOID error __P_((char *));
+extern int _chdir2 __P_((char *));
+extern char *getwd2 __P_((VOID_A));
+extern VOID warning __P_((int, char *));
+#if	MSDOS || !defined (_NODOSDRIVE)
+extern int dospath __P_((char *, char *));
+#endif
+#if	!MSDOS && !defined (_NODOSDRIVE)
+extern int dosstatfs __P_((int, char *));
+#endif
+extern char *strstr2 __P_((char *, char *));
+extern int kanjiputs __P_((char *));
+extern int kanjiputs2 __P_((char *, int, int));
+extern int Xaccess __P_((char *, int));
+extern char *inscomma __P_((char *, long, int));
+extern int kanjiprintf __P_((CONST char *, ...));
 
 extern bindtable bindlist[];
 extern functable funclist[];
 extern char fullpath[];
+extern int sizeinfo;
+extern char *distributor;
 
 #ifndef	MOUNTED
 #define	MOUNTED		"/etc/mtab"
@@ -156,6 +210,9 @@ extern char fullpath[];
 #ifndef	MNTTYPE_UFS
 #define	MNTTYPE_UFS	"ufs"	/* SVR4, OSF/1, FreeBSD, NetBSD */
 #endif
+#ifndef	MNTTYPE_FFS
+#define	MNTTYPE_FFS	"ffs"	/* NetBSD, OpenBSD */
+#endif
 #ifndef	MNTTYPE_EXT2
 #define	MNTTYPE_EXT2	"ext2"	/* Linux */
 #endif
@@ -174,88 +231,49 @@ extern char fullpath[];
 #ifndef	MNTTYPE_PC
 #define	MNTTYPE_PC	"pc"	/* MS-DOS */
 #endif
-
-static int code2str();
-static int checkline();
-static int getfsinfo();
-static char *inscomma();
-static int info1line();
-#ifndef	USEFSDATA
-static long calcKB();
+#ifndef	MNTTYPE_DOS7
+#define	MNTTYPE_DOS7	"dos7"	/* MS-DOS on Win95 */
 #endif
 
+static int code2str __P_((char *, int));
+static int checkline __P_((int));
+static int getfsinfo __P_((char *, statfs_t *, mnt_t *));
+static int info1line __P_((int, char *, long, char *, char *));
+#if	!defined (USEFSDATA) || !defined (_NODOSDRIVE)
+static long calcKB __P_((long, long));
+#endif
+
+static int keycodelist[] = {
+	K_HOME, K_END, K_DL, K_IL, K_DC, K_IC,
+	K_BEG, K_EOL, K_NPAGE, K_PPAGE, K_CLR, K_ENTER,
+	K_BS, '\t', CR, ESC
+};
+static char *keystrlist[] = {
+	"Home", "End", "DelLin", "InsLin", "Del", "Ins",
+	"Beg", "Eol", "PageDn", "PageUp", "Clr", "Enter",
+	"Bs", "Tab", "Ret", "Esc"
+};
 
 static int code2str(buf, code)
 char *buf;
 int code;
 {
-	char *cp;
-
 	buf = buf + strlen(buf);
-	if (code == CR) strcpy(buf, "Ret    ");
-	else if (code == ESC) strcpy(buf, "Esc    ");
-	else if (code == '\t') strcpy(buf, "Tab    ");
-	else if (code < ' ') sprintf(buf, "Ctrl-%c ", code + '@');
-	else if (code < 0x100) sprintf(buf, "'%c'    ", code);
-	else if (code >= K_F(1) && code <= K_F(20))
+	if (code >= K_F(1) && code <= K_F(20))
 		sprintf(buf, "F%-6d", code - K_F0);
+	else if (code == K_UP) sprintf(buf, "%-7.7s", UPAR_K);
+	else if (code == K_DOWN) sprintf(buf, "%-7.7s", DWNAR_K);
+	else if (code == K_RIGHT) sprintf(buf, "%-7.7s", RIGAR_K);
+	else if (code == K_LEFT) sprintf(buf, "%-7.7s", LEFAR_K);
 	else {
-		switch (code) {
-			case K_UP:
-				cp = UPAR_K;
-				break;
-			case K_DOWN:
-				cp = DWNAR_K;
-				break;
-			case K_RIGHT:
-				cp = RIGAR_K;
-				break;
-			case K_LEFT:
-				cp = LEFAR_K;
-				break;
-			case K_HOME:
-				cp = "Home";
-				break;
-			case K_BS:
-				cp = "Bs";
-				break;
-			case K_DL:
-				cp = "DelLin";
-				break;
-			case K_IL:
-				cp = "InsLin";
-				break;
-			case K_DC:
-				cp = "Del";
-				break;
-			case K_IC:
-				cp = "Ins";
-				break;
-			case K_CLR:
-				cp = "Clr";
-				break;
-			case K_EOL:
-				cp = "Eol";
-				break;
-			case K_NPAGE:
-				cp = "PageDn";
-				break;
-			case K_PPAGE:
-				cp = "PageUp";
-				break;
-			case K_ENTER:
-				cp = "Enter";
-				break;
-			case K_BEG:
-				cp = "Beg";
-				break;
-			case K_END:
-				cp = "End";
-				break;
-			default:
-				return(0);
-		}
-		sprintf(buf, "%-7.7s", cp);
+		int i;
+
+		for (i = 0; keycodelist[i]; i++)
+			if (code == keycodelist[i]) break;
+		if (keycodelist[i]) sprintf(buf, "%-7.7s", keystrlist[i]);
+		else if (code < ' ') sprintf(buf, "Ctrl-%c ", code + '@');
+		else if (code < 0x100) sprintf(buf, "'%c'    ", code);
+		else return(0);
 	}
 	return(1);
 }
@@ -278,6 +296,16 @@ int mode;
 	char buf[20];
 	int i, j, c, x, y;
 
+	if (distributor && *distributor) {
+		i = n_column - strlen(distributor) - 26;
+		locate(i, LHELP);
+		putch2('[');
+		putterm(t_standout);
+		cprintf2(" Distributed by: <%s> ", distributor);
+		putterm(end_standout);
+		putch2(']');
+	}
+
 	x = 0;
 	y = WHEADER;
 	locate(0, y++);
@@ -305,10 +333,10 @@ int mode;
 			}
 		if (!c) continue;
 
-		cputs("  ");
-		if (c < 2) cprintf("%-7.7s", " ");
+		cputs2("  ");
+		if (c < 2) cprintf2("%-7.7s", " ");
 		kanjiputs(buf);
-		cputs(": ");
+		cputs2(": ");
 		kanjiputs(mesconv(funclist[i].hmes, funclist[i].hmes_eng));
 
 		y = checkline(y);
@@ -324,6 +352,7 @@ int mode;
 }
 
 #ifdef	USEMNTCTL
+/*ARGSUSED*/
 static FILE *setmntent(file, mode)
 char *file, *mode;
 {
@@ -393,7 +422,8 @@ mnt_t *mntp;
 }
 #endif	/* USEMNTCTL */
 
-#if defined (USEMNTINFOR) || defined (USEMNTINFO) || defined (USEGETFSSTAT)
+#if	defined (USEMNTINFOR) || defined (USEMNTINFO) || defined (USEGETFSSTAT)
+/*ARGSUSED*/
 static FILE *setmntent(file, mode)
 char *file, *mode;
 {
@@ -418,7 +448,7 @@ char *file, *mode;
 	return((FILE *)buf);
 }
 
-#if !defined(MNT_RDONLY) && defined(M_RDONLY)
+#if	!defined (MNT_RDONLY) && defined (M_RDONLY)
 #define	MNT_RDONLY	M_RDONLY
 #endif
 
@@ -435,7 +465,7 @@ mnt_t *mntp;
 	struct vfsconf *conf;
 # define	getvfsbynumber(n)	((conf = getvfsbytype(n)) ? \
 					conf -> vfc_name : NULL)
-# else
+# else	/* !USEVFCNAME */
 #  ifdef	USEFFSTYPE
 #  define	getvfsbynumber(n)	(buf[mnt_ptr].f_fstypename)
 #  else
@@ -447,13 +477,13 @@ mnt_t *mntp;
 #   define	getvfsbynumber(n)	(NULL)
 #   endif
 #  endif
-# endif
-#else	/* USEMNTINFO */
+# endif	/* !USEVFCNAME */
+#else	/* !USEMNTINFO */
 # ifdef	USEGETFSSTAT
 # define	getvfsbynumber(n)	(((n) <= MOUNT_MAXTYPE) ? \
 					mnt_names[n] : NULL)
 # endif
-#endif
+#endif	/* !USEMNTINFO */
 	char *cp;
 	int len;
 
@@ -490,6 +520,7 @@ mnt_t *mntp;
 #endif	/* USEMNTINFOR || USEMNTINFO || USEGETFSSTAT */
 
 #ifdef	USEGETMNT
+/*ARGSUSED*/
 static FILE *setmntent(file, mode)
 char *file, *mode;
 {
@@ -497,6 +528,7 @@ char *file, *mode;
 	return(NULL);
 }
 
+/*ARGSUSED*/
 static mnt_t *getmntent2(fp, mntp)
 FILE *fp;
 mnt_t *mntp;
@@ -529,26 +561,100 @@ mnt_t *mntp;
 
 	return(mntp);
 }
-
-#undef	endmntent
-#define	endmntent(fp)
 #endif	/* USEGETMNT */
 
+#if	MSDOS
+static int statfs2(path, buf)
+char *path;
+statfs_t *buf;
+{
+	union REGS regs;
+	int drv;
+
+	drv = dospath(path, NULL);
+	if (drv >= 'a' && drv <= 'z') drv -= 'a' - 1;
+	else if (drv >= 'A' && drv <= 'Z') drv -= 'A' - 1;
+	else drv = 0;
+
+	regs.x.ax = 0x3600;
+	regs.h.dl = (u_char)drv;
+	intdos(&regs, &regs);
+	if (regs.x.ax == 0xffff) {
+		errno = ENOENT;
+		return(-1);
+	}
+
+	buf -> f_bsize = (long)((u_short)(regs.x.ax) * (u_short)(regs.x.cx));
+	buf -> f_blocks = (u_short)(regs.x.dx);
+	buf -> f_bfree =
+	buf -> f_bavail = (u_short)(regs.x.bx);
+	buf -> f_files = -1;
+
+	return(0);
+}
+#endif
+
 static int getfsinfo(path, fsbuf, mntbuf)
 char *path;
 statfs_t *fsbuf;
 mnt_t *mntbuf;
 {
+#if	MSDOS
+	char *dir;
+
+	mntbuf -> mnt_fsname = "";
+	mntbuf -> mnt_dir[0] = dospath(path, NULL);
+	strcpy(&(mntbuf -> mnt_dir[1]), ":\\");
+	mntbuf -> mnt_type = (supportLFN(mntbuf -> mnt_dir) > 0)
+		? MNTTYPE_DOS7 : MNTTYPE_PC;
+	mntbuf -> mnt_opts = "";
+	dir = path;
+#else	/* !MSDOS */
 	mnt_t *mntp, mnt;
 	FILE *fp;
 	char *dir, fsname[MAXPATHLEN + 1];
 	int len, match;
+#ifndef	_NODOSDRIVE
+	int drv;
+#endif
 
 	dir = NULL;
 	if (!strncmp(path, "/dev/", 4)) {
 		if (_chdir2(path) < 0) dir = path;
 	}
-	else if (dospath(path, NULL) || _chdir2(path) < 0) return(0);
+#ifndef	_NODOSDRIVE
+	else if ((drv = dospath(path, NULL))) {
+		static char dosmntdir[4];
+		char buf[sizeof(long) * 3];
+
+		mntbuf -> mnt_fsname = "";
+		mntbuf -> mnt_dir = dosmntdir;
+		dosmntdir[0] = drv;
+		strcpy(&(dosmntdir[1]), ":\\");
+		mntbuf -> mnt_type = (drv >= 'a' && drv <= 'z')
+			? MNTTYPE_DOS7 : MNTTYPE_PC;
+		mntbuf -> mnt_opts = "";
+		if (dosstatfs(drv, buf) < 0) return(-1);
+		fsbuf -> f_bsize = *((long *)(&buf[sizeof(long) * 0]));
+#ifdef	USEFSDATA
+		fsbuf -> fd_req.btot = calcKB(fsbuf -> f_bsize,
+			*((long *)(&buf[sizeof(long) * 1])));
+		fsbuf -> fd_req.bfree =
+		fsbuf -> fd_req.bfreen = calcKB(fsbuf -> f_bsize,
+			*((long *)(&buf[sizeof(long) * 2])));
+#else
+#ifdef	USESTATVFS
+		fsbuf -> f_frsize = 0;
+#endif
+		fsbuf -> f_blocks = *((long *)(&buf[sizeof(long) * 1]));
+		fsbuf -> f_bfree =
+		fsbuf -> f_bavail = *((long *)(&buf[sizeof(long) * 2]));
+#endif
+		fsbuf -> f_files = -1;
+		return(0);
+	}
+#endif
+	else if (_chdir2(path) < 0) return(-1);
 
 	if (!dir) {
 		dir = getwd2();
@@ -556,31 +662,38 @@ mnt_t *mntbuf;
 		match = 0;
 
 		fp = setmntent(MOUNTED, "r");
-		while (mntp = getmntent2(fp, &mnt)) {
+		while ((mntp = getmntent2(fp, &mnt))) {
 			if ((len = strlen(mntp -> mnt_dir)) < match
 			|| strncmp(mntp -> mnt_dir, dir, len)
-			|| (mntp -> mnt_dir[len - 1] != '/'
-			&& dir[len] && dir[len] != '/')) continue;
+			|| (mntp -> mnt_dir[len - 1] != _SC_
+			&& dir[len] && dir[len] != _SC_)) continue;
 			match = len;
 			strcpy(fsname, mntp -> mnt_fsname);
 		}
 		endmntent(fp);
 
 		free(dir);
-		if (!match) return(0);
+		if (!match) {
+			errno = ENOENT;
+			return(-1);
+		}
 		dir = fsname;
 	}
 	fp = setmntent(MOUNTED, "r");
-	while (mntp = getmntent2(fp, &mnt))
+	while ((mntp = getmntent2(fp, &mnt)))
 		if (!strcmp(dir, mntp -> mnt_fsname)) break;
 	endmntent(fp);
-	if (!mntp) return(0);
+	if (!mntp) {
+		errno = ENOENT;
+		return(-1);
+	}
 	memcpy(mntbuf, mntp, sizeof(mnt_t));
+#endif	/* !MSDOS */
 
 	if (statfs2(mntbuf -> mnt_dir, fsbuf) < 0
 	&& (path == dir || statfs2(path, fsbuf) < 0))
 		memset((char *)fsbuf, 0xff, sizeof(statfs_t));
-	return(1);
+	return(0);
 }
 
 int writablefs(path)
@@ -588,27 +701,50 @@ char *path;
 {
 	statfs_t fsbuf;
 	mnt_t mntbuf;
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	int drv;
+#endif
 
-	if (dospath(path, NULL)) return(0);
-	if (access(path, R_OK | W_OK | X_OK) < 0) return(-1);
-	if (!getfsinfo(path, &fsbuf, &mntbuf)
+	if (Xaccess(path, R_OK | W_OK | X_OK) < 0) return(-1);
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	if ((drv = dospath(path, NULL)))
+		return((drv >= 'A' && drv <= 'Z') ? 4 : 5);
+#endif
+	if (getfsinfo(path, &fsbuf, &mntbuf) < 0
 	|| hasmntopt(&mntbuf, "ro")) return(0);
 
+#if	!MSDOS
 	if (!strcmp(mntbuf.mnt_type, MNTTYPE_43)
 	|| !strcmp(mntbuf.mnt_type, MNTTYPE_42)
 	|| !strcmp(mntbuf.mnt_type, MNTTYPE_UFS)
+	|| !strcmp(mntbuf.mnt_type, MNTTYPE_FFS)
 	|| !strcmp(mntbuf.mnt_type, MNTTYPE_EXT2)
 	|| !strcmp(mntbuf.mnt_type, MNTTYPE_JFS)) return(1);
 	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_EFS)) return(2);
 	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_SYSV)
 	|| !strcmp(mntbuf.mnt_type, MNTTYPE_DGUX)) return(3);
+	else
+#endif
+	if (!strcmp(mntbuf.mnt_type, MNTTYPE_PC)) return(4);
+	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_DOS7)) return(5);
 	return(0);
 }
 
 /*ARGSUSED*/
-int getblocksize(dir)
+long getblocksize(dir)
 char *dir;
 {
+#if	MSDOS
+	statfs_t fsbuf;
+
+	if (statfs2(dir, &fsbuf) < 0) return(1024);
+	return(blocksize(fsbuf));
+#else	/* !MSDOS */
+	statfs_t fsbuf;
+	mnt_t mntbuf;
+
+	if (!strcmp(dir, ".") && getfsinfo(dir, &fsbuf, &mntbuf) >= 0)
+		return(blocksize(fsbuf));
 #ifdef	DEV_BSIZE
 	return(DEV_BSIZE);
 #else
@@ -617,11 +753,13 @@ char *dir;
 	if (stat(dir, &buf) < 0) error(dir);
 	return((int)buf.st_size);
 #endif
+#endif	/* !MSDOS */
 }
 
-static char *inscomma(buf, n, digit)
+char *inscomma(buf, n, digit)
 char *buf;
-int n, digit;
+long n;
+int digit;
 {
 	char tmp[12];
 	int i, j, len;
@@ -629,7 +767,7 @@ int n, digit;
 	j = 0;
 	if (n < 0) buf[j++] = '?';
 	else {
-		sprintf(tmp, "%d", n);
+		sprintf(tmp, "%ld", n);
 		len = strlen(tmp);
 		for (i = j = 0; tmp[i]; i++) {
 			buf[j++] = tmp[i];
@@ -643,7 +781,7 @@ int n, digit;
 static int info1line(y, ind, n, str, unit)
 int y;
 char *ind;
-int n;
+long n;
 char *str, *unit;
 {
 	char buf[16];
@@ -659,13 +797,13 @@ char *str, *unit;
 		kanjiputs2(str, width, 0);
 	}
 	else {
-		cprintf("%12.12s", inscomma(buf, n, 3));
+		cprintf2("%12.12s", inscomma(buf, n, 3));
 		kanjiprintf(" %s", unit);
 	}
 	return(checkline(++y));
 }
 
-#ifndef	USEFSDATA
+#if	!defined (USEFSDATA) || !defined (_NODOSDRIVE)
 static long calcKB(block, byte)
 long block, byte;
 {
@@ -681,6 +819,26 @@ long block, byte;
 }
 #endif
 
+VOID getinfofs(path, totalp, freep)
+char *path;
+long *totalp, *freep;
+{
+	statfs_t fsbuf;
+	mnt_t mntbuf;
+
+	if (getfsinfo(path, &fsbuf, &mntbuf) < 0)
+		*totalp = *freep = -1;
+	else {
+#ifdef	USEFSDATA
+		*totalp = fsbuf.fd_req.btot;
+		*freep = fsbuf.fd_req.bfreen;
+#else
+		*totalp = calcKB(fsbuf.f_blocks, blocksize(fsbuf));
+		*freep = calcKB(fsbuf.f_bavail, blocksize(fsbuf));
+#endif
+	}
+}
+
 int infofs(path)
 char *path;
 {
@@ -688,7 +846,7 @@ char *path;
 	mnt_t mntbuf;
 	int y;
 
-	if (!getfsinfo(path, &fsbuf, &mntbuf)) {
+	if (getfsinfo(path, &fsbuf, &mntbuf) < 0) {
 		warning(ENOTDIR, path);
 		return(0);
 	}
@@ -696,9 +854,9 @@ char *path;
 	locate(0, y++);
 	putterm(l_clear);
 
-	y = info1line(y, FSNAM_K, 0, mntbuf.mnt_fsname, NULL);
-	y = info1line(y, FSMNT_K, 0, mntbuf.mnt_dir, NULL);
-	y = info1line(y, FSTYP_K, 0, mntbuf.mnt_type, NULL);
+	y = info1line(y, FSNAM_K, 0L, mntbuf.mnt_fsname, NULL);
+	y = info1line(y, FSMNT_K, 0L, mntbuf.mnt_dir, NULL);
+	y = info1line(y, FSTYP_K, 0L, mntbuf.mnt_type, NULL);
 #ifdef	USEFSDATA
 	y = info1line(y, FSTTL_K, fsbuf.fd_req.btot, NULL, "Kbytes");
 	y = info1line(y, FSUSE_K,
