@@ -9,16 +9,40 @@
 #include "func.h"
 #include "kctype.h"
 
+#if	MSDOS || defined (__STDC__)
+#include <stdarg.h>
+#else
 #include <varargs.h>
+#endif
 
 #define	ASCII	0
 #define	KANA	1
 #define	KANJI	2
 
-int inputkcode;
-int outputkcode;
+#if	!MSDOS && !defined (_NOKANJICONV)
+int inputkcode = 0;
+#endif
+#if	(!MSDOS && !defined (_NOKANJICONV)) || !defined (_NOENGMES)
+int outputkcode = 0;
+#endif
 
 
+int onkanji1(s, ptr)
+char *s;
+int ptr;
+{
+	int i;
+
+	if (ptr < 0) return(0);
+	if (!ptr) return(iskanji1(s[0]));
+
+	for (i = 0; i < ptr; i++) if (iskanji1(s[i])) i++;
+	if (i > ptr) return(0);
+	return(iskanji1(s[ptr]));
+}
+
+#if	(!MSDOS && !defined (_NOKANJICONV)) || !defined (_NOENGMES)
+/* ARGSUSED */
 int getlang(str, in)
 char *str;
 int in;
@@ -26,13 +50,18 @@ int in;
 	int ret;
 
 	if (!str) ret = NOCNV;
-	else if (strstr(str, "SJIS") || strstr(str, "sjis")) ret = SJIS;
-	else if (strstr(str, "EUC") || strstr(str, "euc")) ret = EUC;
-	else if (strstr(str, "JIS") || strstr(str, "jis")) ret = JIS7;
-	else if (strstr(str, "ENG") || strstr(str, "eng") || !strcmp(str, "C"))
-		ret = ENG;
+#if	!MSDOS && !defined (_NOKANJICONV) 
+	else if (strstr2(str, "SJIS") || strstr2(str, "sjis")) ret = SJIS;
+	else if (strstr2(str, "EUC") || strstr2(str, "euc")) ret = EUC;
+	else if (strstr2(str, "JIS") || strstr2(str, "jis")) ret = JIS7;
+#endif
+#ifndef	_NOENGMES
+	else if (strstr2(str, "ENG") || strstr2(str, "eng")
+	|| !strcmp(str, "C")) ret = ENG;
+#endif
 	else ret = NOCNV;
 
+#if	!MSDOS && !defined (_NOKANJICONV) 
 	if (in) {
 #ifdef	CODEEUC
 		if (ret != SJIS) ret = EUC;
@@ -40,29 +69,20 @@ int in;
 		if (ret != EUC) ret = SJIS;
 #endif
 	}
+#endif	/* !MSDOS && !_NOKANJICONV */
 	return(ret);
 }
+#endif	/* (!MSDOS && !_NOKANJICONV) || !_NOENGMES */
 
+#ifndef	_NOENGMES
 char *mesconv(jpn, eng)
 char *jpn, *eng;
 {
 	return((outputkcode == ENG) ? eng : jpn);
 }
+#endif
 
-int onkanji1(s, ptr)
-u_char *s;
-int ptr;
-{
-	int i;
-
-	if (ptr < 0) return(0);
-	if (!ptr) return(iskanji1((int)s[0]));
-
-	for (i = 0; i < ptr; i++) if (iskanji1((int)s[i])) i++;
-	if (i > ptr) return(0);
-	return(iskanji1((int)s[ptr]));
-}
-
+#if	!MSDOS && !defined (_NOKANJICONV) 
 int jis7(buf, str, incode)
 char *buf;
 u_char *str;
@@ -71,7 +91,7 @@ int incode;
 	int i, j, mode;
 
 	mode = ASCII;
-	for (i = 0, j = 0; str[i] && j < MAXLINESTR - 8; i++, j++) {
+	for (i = j = 0; str[i] && j < MAXLINESTR - 8; i++, j++) {
 		if ((incode == EUC) ? isekana(str, i) : isskana(str, i)) {
 			if (!(mode & KANA)) buf[j++] = '\016';
 			mode |= KANA;
@@ -125,7 +145,7 @@ u_char *str;
 {
 	int i, j;
 
-	for (i = 0, j = 0; str[i] && j < MAXLINESTR; i++, j++) {
+	for (i = j = 0; str[i] && j < MAXLINESTR - 1; i++, j++) {
 		if (isskana(str, i)) {
 			buf[j++] = 0x8e;
 			buf[j] = str[i];
@@ -152,11 +172,11 @@ u_char *str;
 {
 	int i, j;
 
-	for (i = 0, j = 0; str[i] && j < MAXLINESTR; i++, j++) {
+	for (i = j = 0; str[i] && j < MAXLINESTR - 1; i++, j++) {
 		if (isekana(str, i)) buf[j] = str[i];
 		else if (iseuc(str[i])) {
 			buf[j++] = ((str[i] - 0x81) >> 1)
-				+ ((str[i] < 0xde) ? 0x71 : 0xb1);
+				+ ((str[i] < 0xdf) ? 0x71 : 0xb1);
 			buf[j] = str[++i] - 0x80;
 			if (str[i - 1] & 1)
 				buf[j] += (str[i] < 0xe0) ? 0x1f : 0x20;
@@ -168,23 +188,22 @@ u_char *str;
 }
 
 int kanjiconv(buf, str, in, out)
-char *buf;
-u_char *str;
+char *buf, *str;
 int in, out;
 {
 	int len;
 
-	len = strlen((char *)str);
+	len = strlen(str);
 	switch (out) {
 		case JIS7:
-			len = jis7(buf, str, in);
+			len = jis7(buf, (u_char *)str, in);
 			break;
 		case SJIS:
-			if (in == EUC) len = ujis2sjis(buf, str);
+			if (in == EUC) len = ujis2sjis(buf, (u_char *)str);
 			else memcpy(buf, str, len);
 			break;
 		case EUC:
-			if (in == SJIS) len = sjis2ujis(buf, str);
+			if (in == SJIS) len = sjis2ujis(buf, (u_char *)str);
 			else memcpy(buf, str, len);
 			break;
 		default:
@@ -193,20 +212,37 @@ int in, out;
 	}
 	return(len);
 }
+#endif	/* !MSDOS && !_NOKANJICONV */
 
 int kanjiputs(str)
 char *str;
 {
+#if	MSDOS || defined (_NOKANJICONV)
+	cputs2(str);
+	return(strlen(str));
+#else
 	char buf[MAXLINESTR + 1];
 	int i;
 
 	i = kanjiconv(buf, str, DEFCODE, outputkcode);
 	buf[i] = '\0';
-	cputs(buf);
-	return(strlen(buf));
+	cputs2(buf);
+	return(i);
+#endif
 }
 
-#ifndef	NOVSPRINTF
+#if	MSDOS || defined (__STDC__)
+/*VARARGS1*/
+int kanjiprintf(CONST char *fmt, ...)
+{
+	va_list args;
+	char buf[MAXLINESTR + 1];
+
+	va_start(args, fmt);
+	vsprintf(buf, fmt, args);
+	va_end(args);
+#else	/* !MSDOS && !__STDC__ */
+# ifndef	NOVSPRINTF
 /*VARARGS1*/
 int kanjiprintf(fmt, va_alist)
 char *fmt;
@@ -218,14 +254,15 @@ va_dcl
 	va_start(args);
 	vsprintf(buf, fmt, args);
 	va_end(args);
-#else
+# else	/* NOVSPRINTF */
 int kanjiprintf(fmt, arg1, arg2, arg3, arg4, arg5, arg6)
 char *fmt;
 {
 	char buf[MAXLINESTR + 1];
 
 	sprintf(buf, fmt, arg1, arg2, arg3, arg4, arg5, arg6);
-#endif
+# endif	/* NOVSPRINTF */
+#endif	/* !MSDOS && !__STDC__ */
 	return(kanjiputs(buf));
 }
 
@@ -234,20 +271,8 @@ char *s;
 int len, ptr;
 {
 	char dupl[MAXLINESTR + 1];
-	int i, p;
 
-	p = ptr;
-	if (ptr < 0) ptr = 0;
-	for (i = 0; i < len && s[ptr + i]; i++);
-	if (i < len) {
-		kanjiputs(s + ptr);
-		if (p >= 0) for (; i < len; i++) putch(' ');
-	}
-	else {
-		memcpy(dupl, s + ptr, len);
-		dupl[len] = '\0';
-		if (onkanji1((u_char *)dupl, len - 1)) dupl[len - 1] = ' ';
-		kanjiputs(dupl);
-	}
+	len = strncpy3(dupl, s, len, ptr);
+	kanjiputs(dupl);
 	return(len);
 }
