@@ -4,60 +4,154 @@
  *	Terminal Module
  */
 
-#include "machine.h"
 #include <stdio.h>
 #include <string.h>
+#include <ctype.h>
+#include <signal.h>
+#include "machine.h"
+
+#if	defined (DJGPP) && (DJGPP >= 2)
+#include <time.h>
+#endif
+
+#if	MSDOS || defined (__STDC__)
+#include <stdarg.h>
+#else
 #include <varargs.h>
+#endif
+
+#if	MSDOS
+#include <dos.h>
+# ifdef	__GNUC__
+# include <dpmi.h>
+# include <go32.h>
+# include <sys/farptr.h>
+#  if defined (DJGPP) && (DJGPP >= 2)
+#  include <libc/dosio.h>
+#  else
+#  define	_dos_ds		_go32_info_block.selector_for_linear_memory
+#  endif
+# else	/* !__GNUC__ */
+# include <signal.h>
+# include <sys/types.h>
+# include <sys/timeb.h>
+# endif	/* !__GNUC__ */
+#define	TTYNAME		""
+#else	/* !MSDOS */
 #include <fcntl.h>
+#include <errno.h>
 #include <sys/types.h>
 #include <sys/file.h>
 #include <sys/time.h>
 
-#ifndef	NOSTDLIBH
-#include <stdlib.h>
+#ifdef	NOERRNO
+extern int errno;
+#endif
+
+#ifdef	USETERMIOS
+# ifdef	USETERMIO
+# undef	USETERMIO
+# endif
+#include <termios.h>
+typedef	struct termios	termioctl_t;
+#define	tioctl(d, r, a)	((r) ? tcsetattr(d, (r) - 1, a) : tcgetattr(d, a))
+#define	getspeed(t)	cfgetospeed(&t)
+#define	REQGETP		0
+#define	REQSETP		TCSAFLUSH + 1
 #endif
 
 #ifdef	USETERMIO
 #include <termio.h>
-#else
+typedef	struct termio	termioctl_t;
+#define	tioctl		ioctl
+#define	getspeed(t)	((t).c_cflag & CBAUD)
+#define	REQGETP		TCGETA
+#define	REQSETP		TCSETAF
+#endif
+
+#if	!defined (USETERMIO) && !defined(USETERMIOS)
+#define	USESGTTY
 #include <sgtty.h>
+typedef	struct sgttyb	termioctl_t;
+#define	tioctl		ioctl
+#define	getspeed(t)	((t).sg_ospeed)
+#define	REQGETP		TIOCGETP
+#define	REQSETP		TIOCSETP
 #endif
 
 #ifdef	USESELECTH
 #include <sys/select.h>
 #endif
+#endif	/* !MSDOS */
+
+#ifndef	NOSTDLIBH
+#include <stdlib.h>
+#endif
+
+#ifdef	NOVOID
+#define	VOID
+#else
+#define	VOID	void
+#endif
 
 #include "term.h"
 
-extern int tgetent();
-extern int tgetnum();
-extern int tgetflag();
-extern char *tgetstr();
-extern char *tgoto();
-
-#ifndef	SENSEPERSEC
-#define	SENSEPERSEC	50
-#endif
-#ifndef	WAITKEYPAD
-#define	WAITKEYPAD	360		/* msec */
-#endif
 #define	MAXPRINTBUF	255
+#define	SIZEFMT		"\033[%d;%dR"
+
+#if	MSDOS
+#define	GETSIZE		"\033[0;999H\033[999B\033[6n"
+#else	/* !MSDOS */
+extern int tgetent __P_((char *, char *));
+extern int tgetnum __P_((char *));
+extern int tgetflag __P_((char *));
+extern char *tgetstr __P_((char *, char **));
+extern char *tgoto __P_((char *, int, int));
+extern int tputs __P_((char *, int, int (*)__P_((int))));
+
 #define	STDIN		0
 #define	STDOUT		1
 #define	STDERR		2
 #define	TTYNAME		"/dev/tty"
 #define	GETSIZE		"\0337\033[r\033[999;999H\033[6n"
-#define	SIZEFMT		"\033[%d;%dR"
 
 #ifndef	PENDIN
 #define	PENDIN		0
 #endif
+#ifndef	IEXTEN
+#define	IEXTEN		0
+#endif
 #ifndef	ECHOCTL
 #define	ECHOCTL		0
 #endif
 #ifndef	ECHOKE
 #define	ECHOKE		0
 #endif
+#ifndef	OCRNL
+#define	OCRNL		0
+#endif
+#ifndef	ONOCR
+#define	ONOCR		0
+#endif
+#ifndef	ONLRET
+#define	ONLRET		0
+#endif
+#ifndef	TAB3
+#define	TAB3		OXTABS
+#endif
+#ifndef	VINTR
+#define	VINTR		0
+#endif
+#ifndef	VQUIT
+#define	VQUIT		1
+#endif
+#ifndef	VEOF
+#define	VEOF		4
+#endif
+#ifndef	VEOL
+#define	VEOL		5
+#endif
+#define	not(x)		(~(x) & 0xffff)
 
 #ifndef	FREAD
 # ifdef	_FREAD
@@ -74,67 +168,106 @@ typedef struct fd_set {
 # define	FD_ZERO(p)	(((p) -> fds_bits[0]) = 0)
 # define	FD_SET(n, p)	(((p) -> fds_bits[0]) |= (1 << (n)))
 #endif
+#endif	/* !MSDOS */
 
-static int err2();
-static int defaultterm();
-static int tgetstr2();
-static int tgetstr3();
-static int sortkeyseq();
-static int realscanf();
+static int err2 __P_((char *));
+static int defaultterm __P_((VOID_A));
+static int getxy __P_((char *, char *, int *, int *));
+#if	MSDOS
+# ifndef	__GNUC__
+static int dosgettime __P_((u_char []));
+# endif
+#else	/* !MSDOS */
+# ifdef	USESGTTY
+static int ttymode __P_((int, u_short, u_short, u_short, u_short));
+# else
+static int ttymode __P_((int, u_short, u_short, u_short, u_short,
+		u_short, u_short, int, int));
+# endif
+static int tgetstr2 __P_((char **, char *));
+static int tgetstr3 __P_((char **, char *, char *));
+static int sortkeyseq __P_((VOID_A));
+#endif	/* !MSDOS */
 
-short ospeed;
-char PC;
-char *BC;
-char *UP;
-int n_column;
-int n_lastcolumn;
-int n_line;
-int stable_standout;
-char *t_init;
-char *t_end;
-char *t_keypad;
-char *t_nokeypad;
-char *t_normalcursor;
-char *t_highcursor;
-char *t_nocursor;
-char *t_setcursor;
-char *t_resetcursor;
-char *t_bell;
-char *t_vbell;
-char *t_clear;
-char *t_normal;
-char *t_bold;
-char *t_reverse;
-char *t_dim;
-char *t_blink;
-char *t_standout;
-char *t_underline;
-char *end_standout;
-char *end_underline;
-char *l_clear;
-char *l_insert;
-char *l_delete;
-char *c_insert;
-char *c_delete;
-char *c_store;
-char *c_restore;
-char *c_home;
-char *c_locate;
-char *c_return;
-char *c_newline;
-char *c_up;
-char *c_down;
-char *c_right;
-char *c_left;
+short ospeed = 0;
+char PC ='\0';
+char *BC = NULL;
+char *UP = NULL;
+int n_column = 0;
+int n_lastcolumn = 0;
+int n_line = 0;
+int stable_standout = 0;
+char *t_init = NULL;
+char *t_end = NULL;
+char *t_keypad = NULL;
+char *t_nokeypad = NULL;
+char *t_normalcursor = NULL;
+char *t_highcursor = NULL;
+char *t_nocursor = NULL;
+char *t_setcursor = NULL;
+char *t_resetcursor = NULL;
+char *t_bell = NULL;
+char *t_vbell = NULL;
+char *t_clear = NULL;
+char *t_normal = NULL;
+char *t_bold = NULL;
+char *t_reverse = NULL;
+char *t_dim = NULL;
+char *t_blink = NULL;
+char *t_standout = NULL;
+char *t_underline = NULL;
+char *end_standout = NULL;
+char *end_underline = NULL;
+char *l_clear = NULL;
+char *l_insert = NULL;
+char *l_delete = NULL;
+char *c_insert = NULL;
+char *c_delete = NULL;
+char *c_store = NULL;
+char *c_restore = NULL;
+char *c_home = NULL;
+char *c_locate = NULL;
+char *c_return = NULL;
+char *c_newline = NULL;
+char *c_up = NULL;
+char *c_down = NULL;
+char *c_right = NULL;
+char *c_left = NULL;
+u_char cc_intr = CTRL('c');
+u_char cc_quit = CTRL('\\');
+u_char cc_eof = CTRL('d');
+u_char cc_eol = 255;
 
+#if	MSDOS
+#ifdef	PC98
+#define	KEYBUFWORKSEG	0x40
+#define	KEYBUFWORKOFF	0x124
+static int nextchar = '\0';
+static u_char specialkey[] = ":=<;89>\25667bcdefghijk\202\203\204\205\206\207\210\211\212\213?";
+#else
+#define	KEYBUFWORKSEG	0x40
+#define	KEYBUFWORKOFF	0x80
+static u_char specialkey[] = "\003HPMKRSGOIQ;<=>?@ABCDTUVWXYZ[\\]\206";
+#endif
+static int specialkeycode[] = {
+	0,
+	K_UP, K_DOWN, K_RIGHT, K_LEFT,
+	K_IC, K_DC, K_HOME, K_END, K_PPAGE, K_NPAGE,
+	K_F(1), K_F(2), K_F(3), K_F(4), K_F(5),
+	K_F(6), K_F(7), K_F(8), K_F(9), K_F(10),
+	K_F(11), K_F(12), K_F(13), K_F(14), K_F(15),
+	K_F(16), K_F(17), K_F(18), K_F(19), K_F(20), K_HELP
+};
+#else	/* !MSDOS */
 static char *keyseq[K_MAX - K_MIN + 1];
 static u_char keyseqlen[K_MAX - K_MIN + 1];
 static short keycode[K_MAX - K_MIN + 1];
 static int ttyio;
 static FILE *ttyout;
 static char *termname;
+#endif	/* !MSDOS */
 
-#ifndef	TIOCSTI
+#if	MSDOS || !defined (TIOCSTI)
 static u_char ungetbuf[10];
 static int ungetnum = 0;
 #endif
@@ -145,45 +278,150 @@ static int termflags;
 #define	F_INITTERM	004
 
 
+#if	MSDOS
 int inittty(reset)
 int reset;
 {
-#ifdef	USETERMIO
-	static struct termio dupttyio;
-#else
-	static struct sgttyb dupttyio;
-	static int dupttyflag;
-#endif
-	u_long request;
-
-	if (reset && !(termflags & F_INITTTY)) return;
-#ifdef	USETERMIO
-	request = (reset) ? TCSETAW : TCGETA;
-#else
-	request = (reset) ? TIOCLSET : TIOCLGET;
-	if (ioctl(ttyio, request, &dupttyflag) < 0) err2(NULL);
-	request = (reset) ? TIOCSETP : TIOCGETP;
-#endif
-	if (ioctl(ttyio, request, &dupttyio) < 0) err2(NULL);
-	if (!reset) {
-#ifdef	USETERMIO
-		ospeed = dupttyio.c_cflag & CBAUD;
-#else
-		ospeed = dupttyio.sg_ospeed;
-#endif
-		termflags |= F_INITTTY;
-	}
+	if (!reset) termflags |= F_INITTTY;
+	return(0);
 }
 
-#ifdef	USETERMIO
+int cooked2(VOID_A)
+{
+	return(0);
+}
+
+int cbreak2(VOID_A)
+{
+	return(0);
+}
+
+int raw2(VOID_A)
+{
+	return(0);
+}
+
+int echo2(VOID_A)
+{
+	return(0);
+}
+
+int noecho2(VOID_A)
+{
+	return(0);
+}
+
+int nl2(VOID_A)
+{
+	return(0);
+}
+
+int nonl2(VOID_A)
+{
+	return(0);
+}
+
+int tabs(VOID_A)
+{
+	return(0);
+}
+
+int notabs(VOID_A)
+{
+	return(0);
+}
+
+int keyflush(VOID_A)
+{
+#ifdef	__GNUC__
+	u_short keybufp =
+		_farpeekw(_dos_ds, KEYBUFWORKSEG * 0x10 + KEYBUFWORKOFF);
+
+# ifdef	PC98
+	_farpokew(_dos_ds, KEYBUFWORKSEG * 0x10 + KEYBUFWORKOFF + 2, keybufp);
+	_farpokew(_dos_ds, KEYBUFWORKSEG * 0x10 + KEYBUFWORKOFF + 4, 0);
+# else
+	u_short ptr;
+
+	ptr = _farpeekw(_dos_ds, KEYBUFWORKSEG * 0x10 + keybufp - 4);
+	_farpokew(_dos_ds, KEYBUFWORKSEG * 0x10 + keybufp - 2, ptr);
+# endif
+#else	/* !__GNUC__ */
+	u_short far *keybuf = MK_FP(KEYBUFWORKSEG, KEYBUFWORKOFF);
+
+# ifdef	PC98
+	keybuf[2] = 0;
+# else
+	keybuf = MK_FP(KEYBUFWORKSEG, keybuf[0] - 4);
+# endif
+	keybuf[1] = keybuf[0];
+#endif	/* !__GNUC__ */
+	return(0);
+}
+
+#else	/* !MSDOS */
+
+int inittty(reset)
+int reset;
+{
+#ifdef	USESGTTY
+	static int dupttyflag;
+	struct tchars cc;
+#endif
+	static termioctl_t dupttyio;
+	u_long request;
+
+	if (reset && !(termflags & F_INITTTY)) return(0);
+#ifdef	USESGTTY
+	request = (reset) ? TIOCLSET : TIOCLGET;
+	if (tioctl(ttyio, request, &dupttyflag) < 0) err2(NULL);
+#endif
+	request = (reset) ? REQSETP : REQGETP;
+	if (tioctl(ttyio, request, &dupttyio) < 0) err2(NULL);
+
+	if (!reset) {
+#ifdef	USESGTTY
+		tioctl(ttyio, TIOCGETC, &cc);
+		cc_intr = cc.t_intrc;
+		cc_quit = cc.t_quitc;
+		cc_eof = cc.t_eofc;
+		cc_eol = cc.t_brkc;
+#else
+		cc_intr = dupttyio.c_cc[VINTR];
+		cc_quit = dupttyio.c_cc[VQUIT];
+		cc_eof = dupttyio.c_cc[VEOF];
+		cc_eol = dupttyio.c_cc[VEOL];
+#endif
+		ospeed = getspeed(dupttyio);
+		termflags |= F_INITTTY;
+	}
+	return(0);
+}
+
+#ifdef	USESGTTY
+static int ttymode(d, set, reset, lset, lreset)
+int d;
+u_short set, reset, lset, lreset;
+{
+	termioctl_t tty;
+	int lflag;
+
+	if (ioctl(d, TIOCLGET, &lflag) < 0) err2(NULL);
+	if (tioctl(d, REQGETP, &tty) < 0) err2(NULL);
+	if (set) tty.sg_flags |= set;
+	if (reset) tty.sg_flags &= reset;
+	if (set) lflag |= lset;
+	if (lreset) lflag &= lreset;
+	if (ioctl(d, TIOCLSET, &lflag) < 0) err2(NULL);
+#else
 static int ttymode(d, set, reset, iset, ireset, oset, oreset, min, time)
 int d;
 u_short set, reset, iset, ireset, oset, oreset;
 int min, time;
 {
-	struct termio tty;
+	termioctl_t tty;
 
-	if (ioctl(d, TCGETA, &tty) < 0) err2(NULL);
+	if (tioctl(d, REQGETP, &tty) < 0) err2(NULL);
 	if (set) tty.c_lflag |= set;
 	if (reset) tty.c_lflag &= reset;
 	if (iset) tty.c_iflag |= iset;
@@ -194,120 +432,127 @@ int min, time;
 		tty.c_cc[VMIN] = min;
 		tty.c_cc[VTIME] = time;
 	}
-	if (ioctl(d, TCSETAW, &tty) < 0) err2(NULL);
-#else
-static int ttymode(d, set, reset, lset, lreset)
-int d;
-u_short set, reset, lset, lreset;
-{
-	struct sgttyb tty;
-	int lflag;
-
-	if (ioctl(d, TIOCGETP, &tty) < 0) err2(NULL);
-	if (ioctl(d, TIOCLGET, &lflag) < 0) err2(NULL);
-	if (set) tty.sg_flags |= set;
-	if (reset) tty.sg_flags &= reset;
-	if (set) lflag |= lset;
-	if (lreset) lflag &= lreset;
-	if (ioctl(d, TIOCSETP, &tty) < 0) err2(NULL);
-	if (ioctl(d, TIOCLSET, &lflag) < 0) err2(NULL);
 #endif
+	if (tioctl(d, REQSETP, &tty) < 0) err2(NULL);
+	return(0);
 }
 
-int cooked2()
+int cooked2(VOID_A)
 {
-#ifdef	USETERMIO
-	ttymode(ttyio, ISIG | ICANON, ~PENDIN,
-		BRKINT | IGNPAR | IXON | IXANY | IXOFF, ~IGNBRK,
+#ifdef	USESGTTY
+	ttymode(ttyio, 0, not(CBREAK | RAW), LPASS8, not(LLITOUT | LPENDIN));
+#else
+	ttymode(ttyio, ISIG | ICANON | IEXTEN, not(PENDIN),
+		BRKINT | IXON, not(IGNBRK | ISTRIP),
+# if (VEOF == VMIN) || (VEOL == VTIME)
 		OPOST, 0, '\004', 255);
-#else
-	ttymode(ttyio, 0, ~(CBREAK | RAW), 0, ~(LLITOUT | LPENDIN));
+# else
+		OPOST, 0, 0, 0);
+# endif
 #endif
+	return(0);
 }
 
-int cbreak2()
+int cbreak2(VOID_A)
 {
-#ifdef	USETERMIO
-	ttymode(ttyio, ISIG | ICANON, 0, BRKINT, ~(IXON | IGNBRK),
-		0, ~OPOST, 1, 0);
-#else
+#ifdef	USESGTTY
 	ttymode(ttyio, CBREAK, 0, LLITOUT, 0);
+#else
+	ttymode(ttyio, ISIG | IEXTEN, not(ICANON),
+		BRKINT | IXON, not(IGNBRK), OPOST, 0, 1, 0);
 #endif
+	return(0);
 }
 
-int raw2()
+int raw2(VOID_A)
 {
-#ifdef	USETERMIO
-	ttymode(ttyio, 0, ~(ISIG | ICANON), IGNBRK, ~IXON, 0, ~OPOST, 1, 0);
-#else
+#ifdef	USESGTTY
 	ttymode(ttyio, RAW, 0, LLITOUT, 0);
+#else
+	ttymode(ttyio, 0, not(ISIG | ICANON | IEXTEN),
+		IGNBRK, not(BRKINT | IXON), 0, not(OPOST), 1, 0);
 #endif
+	return(0);
 }
 
-int echo2()
+int echo2(VOID_A)
 {
-#ifdef	USETERMIO
-	ttymode(ttyio, ECHO | ECHOE | ECHOK | ECHOCTL | ECHOKE, ~ECHONL,
-		0, 0, 0, 0, 0, 0);
-#else
+#ifdef	USESGTTY
 	ttymode(ttyio, ECHO, 0, LCRTBS | LCRTERA | LCRTKIL | LCTLECH, 0);
+#else
+	ttymode(ttyio, ECHO | ECHOE | ECHOCTL | ECHOKE, not(ECHONL),
+		0, 0, 0, 0, 0, 0);
 #endif
+	return(0);
 }
 
-int noecho2()
+int noecho2(VOID_A)
 {
-#ifdef	USETERMIO
-	ttymode(ttyio, 0, ~(ECHO | ECHOE | ECHOK | ECHONL), 0, 0, 0, 0, 0, 0);
+#ifdef	USESGTTY
+	ttymode(ttyio, 0, not(ECHO), 0, not(LCRTBS | LCRTERA));
 #else
-	ttymode(ttyio, 0, ~ECHO, 0, ~(LCRTBS | LCRTERA));
+	ttymode(ttyio, 0, not(ECHO | ECHOE | ECHOK | ECHONL), 0, 0,
+		0, 0, 0, 0);
 #endif
+	return(0);
 }
 
-int nl2()
+int nl2(VOID_A)
 {
-#ifdef	USETERMIO
-	ttymode(ttyio, 0, 0, ICRNL, 0, ONLCR, ~(OCRNL | ONOCR | ONLRET), 0, 0);
-#else
+#ifdef	USESGTTY
 	ttymode(ttyio, CRMOD, 0, 0, 0);
+#else
+	ttymode(ttyio, 0, 0, ICRNL, 0,
+		ONLCR, not(OCRNL | ONOCR | ONLRET), 0, 0);
 #endif
+	return(0);
 }
 
-int nonl2()
+int nonl2(VOID_A)
 {
-#ifdef	USETERMIO
-	ttymode(ttyio, 0, 0, 0, ~ICRNL, 0, ~ONLCR, 0, 0);
+#ifdef	USESGTTY
+	ttymode(ttyio, 0, not(CRMOD), 0, 0);
 #else
-	ttymode(ttyio, 0, ~CRMOD, 0, 0);
+	ttymode(ttyio, 0, 0, 0, not(ICRNL), 0, not(ONLCR), 0, 0);
 #endif
+	return(0);
 }
 
-int tabs()
+int tabs(VOID_A)
 {
-#ifdef	USETERMIO
-	ttymode(ttyio, 0, 0, 0, 0, 0, ~TAB3, 0, 0);
+#ifdef	USESGTTY
+	ttymode(ttyio, 0, not(XTABS), 0, 0);
 #else
-	ttymode(ttyio, 0, ~XTABS, 0, 0);
+	ttymode(ttyio, 0, 0, 0, 0, 0, not(TAB3), 0, 0);
 #endif
+	return(0);
 }
 
-int notabs()
+int notabs(VOID_A)
 {
-#ifdef	USETERMIO
-	ttymode(ttyio, 0, 0, 0, 0, TAB3, 0, 0, 0);
-#else
+#ifdef	USESGTTY
 	ttymode(ttyio, XTABS, 0, 0, 0);
+#else
+	ttymode(ttyio, 0, 0, 0, 0, TAB3, 0, 0, 0);
 #endif
+	return(0);
 }
 
-int keyflush()
+int keyflush(VOID_A)
 {
-#ifdef	USETERMIO
-	ioctl(ttyio, TCFLSH, 0);
-#else
+#ifdef	USESGTTY
 	int i = FREAD;
-	ioctl(ttyio, TIOCFLUSH, &i);
-#endif
+	tioctl(ttyio, TIOCFLUSH, &i);
+#else	/* !USESGTTY */
+# ifdef	USETERMIOS
+	tcflush(ttyio, TCIFLUSH);
+# else
+	tioctl(ttyio, TCFLSH, 0);
+# endif
+#endif	/* !USESGTTY */
+	return(0);
 }
+#endif	/* !MSDOS */	
 
 int exit2(no)
 int no;
@@ -317,6 +562,7 @@ int no;
 	inittty(1);
 	keyflush();
 	exit(no);
+	return(0);
 }
 
 static int err2(mes)
@@ -328,19 +574,37 @@ char *mes;
 	fprintf(stderr, "\007\n");
 	if (mes) fprintf(stderr, "%s\n", mes);
 	else perror(TTYNAME);
-	exit(127);
+	exit(1);
+	return(0);
 }
 
-static int defaultterm()
+static int defaultterm(VOID_A)
 {
+#if	!MSDOS
 	int i;
+#endif
 
 	BC = "\010";
-	UP = "2\033[A";
-	n_column = n_lastcolumn = 80;
+	UP = "\033[A";
+	n_column = 80;
+#if	MSDOS
+	n_lastcolumn = 79;
+	n_line = 25;
+#else
+	n_lastcolumn = 80;
 	n_line = 24;
+#endif
 	t_init = "";
 	t_end = "";
+#if	MSDOS
+	t_keypad = "";
+	t_nokeypad = "";
+	t_normalcursor = "\033[>5l";
+	t_highcursor = "\033[>5l";
+	t_nocursor = "\033[>5h";
+	t_setcursor = "\033[s";
+	t_resetcursor = "\033[u";
+#else
 	t_keypad = "\033[?1h\033=";
 	t_nokeypad = "\033[?1l\033>";
 	t_normalcursor = "\033[?25h";
@@ -348,19 +612,20 @@ static int defaultterm()
 	t_nocursor = "\033[?25l";
 	t_setcursor = "\0337";
 	t_resetcursor = "\0338";
+#endif
 	t_bell = "\007";
 	t_vbell = "\007";
-	t_clear = "50\033[;H\033[2J";
-	t_normal = "2\033[m";
-	t_bold = "2\033[1m";
-	t_reverse = "2\033[7m";
-	t_dim = "2\033[2m";
-	t_blink = "2\033[5m";
-	t_standout = "2\033[7m";
-	t_underline = "2\033[4m";
-	end_standout = "2\033[m";
-	end_underline = "2\033[m";
-	l_clear = "3\033[K";
+	t_clear = "\033[;H\033[2J";
+	t_normal = "\033[m";
+	t_bold = "\033[1m";
+	t_reverse = "\033[7m";
+	t_dim = "\033[2m";
+	t_blink = "\033[5m";
+	t_standout = "\033[7m";
+	t_underline = "\033[4m";
+	end_standout = "\033[m";
+	end_underline = "\033[m";
+	l_clear = "\033[K";
 	l_insert = "\033[L";
 	l_delete = "\033[M";
 	c_insert = "";
@@ -368,20 +633,25 @@ static int defaultterm()
 	c_store = "\033[s";
 	c_restore = "\033[u";
 	c_home = "\033[H";
-	c_locate = "5\033[%i%d;%dH";
+#ifdef	MSDOS
+	c_locate = "\033[%d;%dH";
+#else
+	c_locate = "\033[%i%d;%dH";
+#endif
 	c_return = "\r";
 	c_newline = "\n";
-	c_up = "2\033[A";
+	c_up = "\033[A";
 	c_down = "\012";
-	c_right = "2\033[C";
+	c_right = "\033[C";
 	c_left = "\010";
 
+#if	!MSDOS
 	for (i = 0; i <= K_MAX - K_MIN; i++) keyseq[i] = NULL;
 	keyseq[K_UP - K_MIN] = "\033OA";
 	keyseq[K_DOWN - K_MIN] = "\033OB";
 	keyseq[K_RIGHT - K_MIN] = "\033OC";
 	keyseq[K_LEFT - K_MIN] = "\033OD";
-	keyseq[K_HOME - K_MIN] = "\033[1~";
+	keyseq[K_HOME - K_MIN] = "\033[4~";
 	keyseq[K_BS - K_MIN] = "\010";
 	keyseq[K_F(1) - K_MIN] = "\033[11~";
 	keyseq[K_F(2) - K_MIN] = "\033[12~";
@@ -440,9 +710,46 @@ static int defaultterm()
 	keyseq[K_PPAGE - K_MIN] = "\033[5~";
 	keyseq[K_NPAGE - K_MIN] = "\033[6~";
 	keyseq[K_ENTER - K_MIN] = "\033[9~";
-	keyseq[K_END - K_MIN] = "\033[4~";
+	keyseq[K_END - K_MIN] = "\033[1~";
+#endif
+	return(0);
 }
 
+static int getxy(s, format, yp, xp)
+char *s, *format;
+int *yp, *xp;
+{
+	char *cp;
+	int i, j, tmp, count, *val[2];
+
+	count = 0;
+	val[0] = yp;
+	val[1] = xp;
+
+	for (i = 0, j = 0; s[i] && format[j]; i++, j++) {
+		if (format[j] == '%' && format[++j] == 'd') {
+			for (cp = &s[i]; *cp && *cp >= '0' && *cp <= '9'; cp++);
+			if (cp == &s[i]) break;
+			tmp = *cp;
+			*cp = '\0';
+			*val[count++] = atoi(&s[i]);
+			*cp = tmp;
+		}
+		else if (s[i] != format[j]) j--;
+	}
+	return(count);
+}
+
+#if	MSDOS
+int getterment(VOID_A)
+{
+	defaultterm();
+	termflags |= F_TERMENT;
+	return(0);
+}
+
+#else	/* !MSDOS */
+
 static int tgetstr2(term, str)
 char **term;
 char *str;
@@ -451,10 +758,11 @@ char *str;
 	char *p, *cp;
 
 	p = strbuf;
-	if (cp = tgetstr(str, &p)) {
+	if ((cp = tgetstr(str, &p)) || (cp = *term)) {
 		if (!(*term = (char *)malloc(strlen(cp) + 1))) err2(NULL);
 		strcpy(*term, cp);
 	}
+	return(0);
 }
 
 static int tgetstr3(term, str1, str2)
@@ -471,35 +779,14 @@ char *str1, *str2;
 		if (!(*term = (char *)malloc(strlen(cp) + 1))) err2(NULL);
 		strcpy(*term, cp);
 	}
+	return(0);
 }
 
-static int sortkeyseq()
+static int sortkeyseq(VOID_A)
 {
 	int i, j, tmp;
 	char *str;
 
-	for (i = 0; i <= K_MAX - K_MIN; i++) keycode[i] = K_MIN + i;
-	for (i = 1; i <= 10; i++) keycode[K_F(i + 20) - K_MIN] = K_F(i);
-
-	keycode[K_F('*') - K_MIN] = '*';
-	keycode[K_F('+') - K_MIN] = '+';
-	keycode[K_F(',') - K_MIN] = ',';
-	keycode[K_F('-') - K_MIN] = '-';
-	keycode[K_F('.') - K_MIN] = '.';
-	keycode[K_F('/') - K_MIN] = '/';
-	keycode[K_F('0') - K_MIN] = '0';
-	keycode[K_F('1') - K_MIN] = '1';
-	keycode[K_F('2') - K_MIN] = '2';
-	keycode[K_F('3') - K_MIN] = '3';
-	keycode[K_F('4') - K_MIN] = '4';
-	keycode[K_F('5') - K_MIN] = '5';
-	keycode[K_F('6') - K_MIN] = '6';
-	keycode[K_F('7') - K_MIN] = '7';
-	keycode[K_F('8') - K_MIN] = '8';
-	keycode[K_F('9') - K_MIN] = '9';
-	keycode[K_F('=') - K_MIN] = '=';
-	keycode[K_F('?') - K_MIN] = CR;
-
 	for (i = 0; i < K_MAX - K_MIN; i++)
 		for (j = i + 1; j <= K_MAX - K_MIN; j++)
 			if (!keyseq[j]
@@ -514,11 +801,13 @@ static int sortkeyseq()
 
 	for (i = 0; i <= K_MAX - K_MIN; i++)
 		keyseqlen[i] = (keyseq[i]) ? strlen(keyseq[i]) : 0;
+	return(0);
 }
 
-int getterment()
+int getterment(VOID_A)
 {
 	char *cp, buf[1024];
+	int i;
 
 	if (!(termname = (char *)getenv("TERM"))) termname = "unknown";
 	if (tgetent(buf, termname) <= 0) err2("No TERMCAP prepared");
@@ -611,68 +900,315 @@ int getterment()
 	tgetstr2(&keyseq[K_ENTER - K_MIN], "@8");
 	tgetstr2(&keyseq[K_BEG - K_MIN], "@1");
 	tgetstr2(&keyseq[K_END - K_MIN], "@7");
+
+	for (i = 0; i <= K_MAX - K_MIN; i++) keycode[i] = K_MIN + i;
+	for (i = 1; i <= 10; i++) keycode[K_F(i + 20) - K_MIN] = K_F(i);
+
+	keycode[K_F('*') - K_MIN] = '*';
+	keycode[K_F('+') - K_MIN] = '+';
+	keycode[K_F(',') - K_MIN] = ',';
+	keycode[K_F('-') - K_MIN] = '-';
+	keycode[K_F('.') - K_MIN] = '.';
+	keycode[K_F('/') - K_MIN] = '/';
+	keycode[K_F('0') - K_MIN] = '0';
+	keycode[K_F('1') - K_MIN] = '1';
+	keycode[K_F('2') - K_MIN] = '2';
+	keycode[K_F('3') - K_MIN] = '3';
+	keycode[K_F('4') - K_MIN] = '4';
+	keycode[K_F('5') - K_MIN] = '5';
+	keycode[K_F('6') - K_MIN] = '6';
+	keycode[K_F('7') - K_MIN] = '7';
+	keycode[K_F('8') - K_MIN] = '8';
+	keycode[K_F('9') - K_MIN] = '9';
+	keycode[K_F('=') - K_MIN] = '=';
+	keycode[K_F('?') - K_MIN] = CR;
+
 	sortkeyseq();
 
 	termflags |= F_TERMENT;
+	return(0);
 }
 
-int initterm()
+int setkeyseq(n, str)
+int n;
+char *str;
+{
+	int i;
+
+	for (i = 0; i <= K_MAX - K_MIN; i++) if (keycode[i] == n) {
+		if (keyseq[i]) free(keyseq[i]);
+		keyseq[i] = str;
+		break;
+	}
+	if (i > K_MAX - K_MIN) return(-1);
+
+	if (str) for (i = 0; i <= K_MAX - K_MIN; i++) {
+		if (keycode[i] == n || !keyseq[i] || strcmp(keyseq[i], str))
+			continue;
+		free(keyseq[i]);
+		keyseq[i] = NULL;
+	}
+	sortkeyseq();
+	return(0);
+}
+
+char *getkeyseq(n)
+int n;
+{
+	int i;
+
+	for (i = 0; i <= K_MAX - K_MIN; i++)
+		if (keycode[i] == n) return(keyseq[i]);
+	return(NULL);
+}
+#endif	/* !MSDOS */
+
+int initterm(VOID_A)
 {
 	if (!(termflags & F_TERMENT)) getterment();
 	putterms(t_keypad);
 	putterms(t_init);
 	tflush();
 	termflags |= F_INITTERM;
+	return(0);
 }
 
-int endterm()
+int endterm(VOID_A)
 {
-	if (!(termflags & F_INITTERM)) return;
+	if (!(termflags & F_INITTERM)) return(0);
 	putterms(t_nokeypad);
 	putterms(t_end);
 	tflush();
 	termflags &= ~F_INITTERM;
+	return(0);
 }
 
-int putch(c)
+#if	MSDOS
+
+int putch2(c)
+int c;
+{
+	bdos(0x06, c & 0xff, 0);
+	return(c);
+}
+
+int cputs2(s)
+char *s;
+{
+	char *size, buf[sizeof(SIZEFMT) + 4];
+	int i, x, y;
+
+	for (i = 0; s[i]; i++) {
+		if (s[i] != '\t') {
+			putch2(s[i]);
+			continue;
+		}
+		putch2('\033');
+		putch2('[');
+		putch2('6');
+		putch2('n');
+		size = SIZEFMT;
+			
+		for (x = 0; (buf[x] = bdos(0x07, 0x00, 0)) != '\r'; x++);
+		buf[x] = '\0';
+		if (getxy(buf, size, &y, &x) != 2) x = 1;
+		do {
+			putch2(' ');
+		} while (x++ % 8);
+	}
+	return(0);
+}
+
+/*ARGSUSED*/
+int kbhit2(usec)
+u_long usec;
+{
+#ifdef	PC98
+	union REGS regs;
+
+	if (nextchar) return(1);
+
+	regs.h.ah = 0x01;
+	int86(0x18, &regs, &regs);
+	return(regs.h.bh != 0);
+#else
+# if	defined (DJGPP) && (DJGPP >= 2)
+	fd_set readfds;
+	struct timeval timeout;
+
+	timeout.tv_sec = 0L;
+	timeout.tv_usec = usec;
+	FD_ZERO(&readfds);
+	FD_SET(0, &readfds);
+
+	return(select(1, &readfds, NULL, NULL, &timeout));
+# else	/* !DJGPP || DJGPP < 2 */
+	union REGS regs;
+
+	regs.h.ah = 0x01;
+	int86(0x16, &regs, &regs);
+	return((regs.x.flags & 0x40) ? 0 : 1);
+# endif
+#endif
+}
+
+int getch2(VOID_A)
+{
+#ifdef	PC98
+	union REGS regs;
+	int ch;
+
+	if (nextchar) {
+		ch = nextchar;
+		nextchar = '\0';
+		return(ch);
+	}
+#endif
+	if (ungetnum > 0) return((int)ungetbuf[--ungetnum]);
+
+#ifdef	PC98
+	regs.h.ah = 0x00;
+	int86(0x18, &regs, &regs);
+
+	if (!(ch = regs.h.al)) nextchar = regs.h.ah;
+	return(ch);
+#else
+	return(bdos(0x07, 0x00, 0) & 0xff);
+#endif
+}
+
+#ifndef	__GNUC__
+static int dosgettime(tbuf)
+u_char tbuf[];
+{
+	union REGS regs;
+
+	regs.x.ax = 0x2c00;
+	intdos(&regs, &regs);
+	tbuf[0] = regs.h.ch;
+	tbuf[1] = regs.h.cl;
+	tbuf[2] = regs.h.dh;
+	return(0);
+}
+#endif
+
+/*ARGSUSED*/
+int getkey2(sig)
+int sig;
+{
+#ifndef	__GNUC__
+	static u_char tbuf1[3] = {0xff, 0xff, 0xff};
+	u_char tbuf2[3];
+#endif
+#if	defined (DJGPP) && (DJGPP >= 2)
+	static int count = SENSEPERSEC;
+#endif
+	int i, ch;
+
+#ifdef	__GNUC__
+	do {
+		i = kbhit2(1000000L / SENSEPERSEC);
+# if	defined (DJGPP) && (DJGPP >= 2)
+		if (sig && !(--count)) {
+			count = SENSEPERSEC;
+			raise(sig);
+		}
+# endif
+	} while (!i);
+#else
+	if (tbuf1[0] == 0xff) dosgettime(tbuf1);
+	while (!kbhit2(1000000L / SENSEPERSEC)) {
+		dosgettime(tbuf2);
+		if (memcmp(tbuf1, tbuf2, sizeof(tbuf1))) {
+			if (sig) raise(sig);
+			memcpy(tbuf1, tbuf2, sizeof(tbuf1));
+		}
+	}
+#endif
+	ch = getch2();
+
+	if (ch) switch (ch) {
+		case '\010':
+			ch = K_BS;
+			break;
+		case '\177':
+			ch = K_DC;
+			break;
+		default:
+			break;
+	}
+	else
+#if	defined (PC98) || (defined (DJGPP) && (DJGPP >= 2))
+	if (kbhit2(WAITKEYPAD * 1000L))
+#endif
+	{
+		ch = getch2();
+		for (i = 0; i < sizeof(specialkey) / sizeof(u_char); i++)
+			if (ch == specialkey[i]) return(specialkeycode[i]);
+		ch = K_NOKEY;
+	}
+	return(ch);
+}
+
+int ungetch2(c)
+u_char c;
+{
+	if (ungetnum >= sizeof(ungetbuf) / sizeof(u_char) - 1) return(EOF);
+	ungetbuf[ungetnum++] = c;
+	return(c);
+}
+
+int locate(x, y)
+int x, y;
+{
+	cprintf2(c_locate, ++y, ++x);
+	return(0);
+}
+
+int tflush(VOID_A)
+{
+	return(0);
+}
+
+int getwsize(xmax, ymax)
+int xmax, ymax;
+{
+	char *size, buf[sizeof(SIZEFMT) + 4];
+	int i, x, y;
+
+	size = SIZEFMT;
+
+	for (i = 0; i < sizeof(GETSIZE) - 1; i++) putch2(GETSIZE[i]);
+	for (i = 0; (buf[i] = bdos(0x07, 0x00, 0)) != '\r'; i++);
+	buf[i] = '\0';
+	if (getxy(buf, size, &y, &x) != 2) x = y = -1;
+
+	if (x > 0) {
+		n_lastcolumn = (n_lastcolumn < n_column) ? x - 1 : x;
+		n_column = x;
+	}
+	if (y > 0) n_line = y;
+
+	if (n_column < xmax) err2("Column size too small");
+	if (n_line < ymax) err2("Line size too small");
+
+	return(0);
+}
+
+#else	/* !MSDOS */
+
+int putch2(c)
 int c;
 {
 	return(fputc(c, ttyout));
 }
 
-int cputs(str)
+int cputs2(str)
 char *str;
 {
 	return(fputs(str, ttyout));
 }
 
-#ifndef	NOVSPRINTF
-/*VARARGS1*/
-int cprintf(fmt, va_alist)
-char *fmt;
-va_dcl
-{
-	va_list args;
-	int len;
-	char buf[MAXPRINTBUF + 1];
-
-	va_start(args);
-	len = vsprintf(buf, fmt, args);
-	va_end(args);
-#else
-int cprintf(fmt, arg1, arg2, arg3, arg4, arg5, arg6)
-char *fmt;
-{
-	int len;
-	char buf[MAXPRINTBUF + 1];
-
-	len = sprintf(buf, fmt, arg1, arg2, arg3, arg4, arg5, arg6);
-#endif
-
-	fputs(buf, ttyout);
-	return(len);
-}
-
 int kbhit2(usec)
 u_long usec;
 {
@@ -691,19 +1227,21 @@ u_long usec;
 #endif
 }
 
-int getch2()
+int getch2(VOID_A)
 {
 	u_char ch;
+	int i;
 
-	read(ttyio, &ch, sizeof(u_char));
+	while ((i = read(ttyio, &ch, sizeof(u_char))) < 0 && errno == EINTR);
+	if (i < sizeof(u_char)) return(EOF);
 	return((int)ch);
 }
 
-int getkey(sig)
+int getkey2(sig)
 int sig;
 {
 	static int count = SENSEPERSEC;
-	int i, j, ch, key, start;
+	int i, j, ch, key, start, end, s, e;
 
 	do {
 		key = kbhit2(1000000L / SENSEPERSEC);
@@ -720,15 +1258,23 @@ int sig;
 	} while (!key);
 
 	key = ch = getch2();
-	for (i = start = 0; ; i++) {
-		for (j = start, start = -1; j <= K_MAX - K_MIN; j++) {
+	start = 0;
+	end = K_MAX - K_MIN;
+
+	for (i = 0; ; i++) {
+		s = start;
+		e = end;
+		start = end = -1;
+		for (j = s; j <= e; j++) {
 			if (keyseqlen[j] > i && ch == keyseq[j][i]) {
 				if (keyseqlen[j] == i + 1) return(keycode[j]);
-				if (start < 0) start = j;
+				end = j;
 			}
-			else if (start >= 0) break;
+			else if (end < 0) start = j + 1;
+			else break;
 		}
-		if (start < 0 || !kbhit2(WAITKEYPAD * 1000L)) return(key);
+		if (start < 0) start = s;
+		if (start > end || !kbhit2(WAITKEYPAD * 1000L)) return(key);
 		ch = getch2();
 	}
 }
@@ -739,45 +1285,23 @@ u_char c;
 #ifdef	TIOCSTI
 	ioctl(ttyio, TIOCSTI, &c);
 #else
-	ungetbuf[ungetnum] = c;
-	if (ungetnum < sizeof(ungetbuf) - 1) ungetnum++;
+	if (ungetnum >= sizeof(ungetbuf) / sizeof(u_char) - 1) return(EOF);
+	ungetbuf[ungetnum++] = c;
 #endif
+	return(c);
 }
 
 int locate(x, y)
 int x, y;
 {
 	putterms(tgoto(c_locate, x, y));
+	return(0);
 }
 
-int tflush()
+int tflush(VOID_A)
 {
 	fflush(ttyout);
-}
-
-static int realscanf(s, format, yp, xp)
-char *s, *format;
-int *yp, *xp;
-{
-	char *cp;
-	int i, j, tmp, count, *val[2];
-
-	count = 0;
-	val[0] = yp;
-	val[1] = xp;
-
-	for (i = 0, j = 0; s[i] && format[j]; i++, j++) {
-		if (format[j] == '%' && format[++j] == 'd') {
-			for (cp = &s[i]; *cp && *cp >= '0' && *cp <= '9'; cp++);
-			if (cp == &s[i]) break;
-			tmp = *cp;
-			*cp = '\0';
-			*val[count++] = atoi(&s[i]);
-			*cp = tmp;
-		}
-		else if (s[i] != format[j]) j--;
-	}
-	return(count);
+	return(0);
 }
 
 int getwsize(xmax, ymax)
@@ -824,9 +1348,7 @@ int xmax, ymax;
 		buf[++i] = '\0';
 		fclose(fp);
 
-		if (realscanf(buf, size, &y, &x) != 2) {
-			x = y = -1;
-		}
+		if (getxy(buf, size, &y, &x) != 2) x = y = -1;
 	}
 
 	if (x > 0) {
@@ -837,4 +1359,54 @@ int xmax, ymax;
 
 	if (n_column < xmax) err2("Column size too small");
 	if (n_line < ymax) err2("Line size too small");
+
+	return(0);
+}
+#endif	/* !MSDOS */
+
+#if	MSDOS || defined (__STDC__)
+/*VARARGS1*/
+int cprintf2(CONST char *fmt, ...)
+{
+	va_list args;
+	char buf[MAXPRINTBUF + 1];
+
+	va_start(args, fmt);
+	vsprintf(buf, fmt, args);
+	va_end(args);
+#else	/* !MSDOS && !__STDC__ */
+# ifndef	NOVSPRINTF
+/*VARARGS1*/
+int cprintf2(fmt, va_alist)
+char *fmt;
+va_dcl
+{
+	va_list args;
+	char buf[MAXPRINTBUF + 1];
+
+	va_start(args);
+	vsprintf(buf, fmt, args);
+	va_end(args);
+# else	/* NOVSPRINTF */
+int cprintf2(fmt, arg1, arg2, arg3, arg4, arg5, arg6)
+char *fmt;
+{
+	char buf[MAXPRINTBUF + 1];
+
+	sprintf(buf, fmt, arg1, arg2, arg3, arg4, arg5, arg6);
+# endif	/* NOVSPRINTF */
+#endif	/* !MSDOS && !__STDC__ */
+	return(cputs2(buf));
+}
+
+int chgcolor(color, reverse)
+int color, reverse;
+{
+	if (!reverse) cprintf2("\033[%dm", color + ANSI_NORMAL);
+	else if (color == ANSI_BLACK)
+		cprintf2("\033[%dm\033[%dm",
+			ANSI_WHITE + ANSI_NORMAL, color + ANSI_REVERSE);
+	else cprintf2("\033[%dm\033[%dm",
+			ANSI_BLACK + ANSI_NORMAL, color + ANSI_REVERSE);
+	return(0);
 }
